"""Meeting data structure and related functionality."""

import enum
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, Optional

from ..agents.base_agent import BaseAgent
from ..message import Message


class MeetingInvitationStatus(enum.Enum):
    PENDING = "pending"
    ACCEPTED = "accepted"
    REJECTED = "rejected"


@dataclass
class MeetingInvitation:
    """Represents an invitation to a meeting."""

    agent: BaseAgent
    created_at: datetime
    status: MeetingInvitationStatus = MeetingInvitationStatus.PENDING
    resolved_at: Optional[datetime] = None


@dataclass
class Meeting:
    """Represents an active meeting."""

    id: str
    created_at: datetime
    owner_id: str
    topic: Optional[str] = None

    required_attendees: List[BaseAgent] = field(default_factory=list)
    optional_attendees: List[BaseAgent] = field(default_factory=list)
    joined_attendees: List[BaseAgent] = field(default_factory=list)
    invitations: Dict[str, MeetingInvitation] = field(default_factory=dict)

    message_history: List["Message"] = field(
        default_factory=list
    )  # All messages in this meeting
    agent_last_message_index: Dict[str, int] = field(default_factory=dict)

    def agent_joined(self, agent: BaseAgent) -> None:
        """Add a participant to the meeting."""
        self.joined_attendees.append(agent)
        invitation = self.invitations.get(agent.id)
        if invitation:
            invitation.status = MeetingInvitationStatus.ACCEPTED
            invitation.resolved_at = datetime.now()

    def agent_rejected(self, agent: BaseAgent) -> None:
        invitation = self.invitations.get(agent.id)
        if invitation:
            invitation.status = MeetingInvitationStatus.REJECTED
            invitation.resolved_at = datetime.now()

    def agent_left(self, agent: BaseAgent) -> None:
        """Remove a participant from the meeting."""
        self.joined_attendees.remove(agent)
        self.agent_last_message_index.pop(agent.id, None)

    def has_pending_invitations(self) -> bool:
        """Check if there are any pending invitations."""
        return any(
            invitation.status == MeetingInvitationStatus.PENDING
            for invitation in self.invitations.values()
        )

    def missing_required_attendees(self) -> List[BaseAgent]:
        """Get the list of required attendees that are not present."""
        return [
            attendee
            for attendee in self.required_attendees
            if attendee not in self.joined_attendees
        ]

    def log_message(self, message: "Message") -> None:
        """Add a message to the meeting history."""
        self.message_history.append(message)

    def get_unread_messages(self, agent: BaseAgent) -> List["Message"]:
        """Get unread messages for a specific agent."""
        last_index = self.agent_last_message_index.get(agent.id, 0)
        return self.message_history[last_index:]

    def mark_messages_read(self, agent: BaseAgent) -> None:
        """Mark all messages as read for a specific agent."""
        self.agent_last_message_index[agent.id] = len(self.message_history)

    def is_participant(self, agent: BaseAgent) -> bool:
        """Check if an agent is a participant in the meeting."""
        return agent in self.joined_attendees

    def has_pending_invitation(self, agent: BaseAgent) -> bool:
        """Check if an agent has a pending invitation."""
        return (
            agent.id in self.invitations
            and self.invitations[agent.id].status == MeetingInvitationStatus.PENDING
        )


@dataclass
class JoinedMeeting:
    """Represents a meeting that an agent has joined."""

    id: str
    owner_id: str
    joined_at: datetime



"""
Zero-overhead debug logging for Playbooks framework development.

Usage:
    from playbooks.debug_logger import debug

    # Basic usage
    debug("Agent processing message", agent_id="1234", message_type="USER_INPUT")
    debug("Performance: operation took {duration:.2f}ms", duration=15.5)

    # With Rich markup for colored output
    debug("[green]Success![/green] Operation completed")
    debug("Status: [bold red]ERROR[/bold red] - [yellow]Retrying...[/yellow]")
    debug("[cyan]Info:[/cyan] Processing [bold magenta]agent-123[/bold magenta]")

    # Disable markup parsing if needed
    debug("Literal [brackets] in message", markup=False)

    # Custom colors via context
    debug("Custom colored message", color="purple")
    debug("Bold and italic", style="bold italic blue")
"""

import logging
import os
from typing import Any, Optional

from rich.console import Console
from rich.text import Text

from .config import config
from .logging_constants import (
    DEBUG_LOGGER_NAME,
    DEBUG_PREFIX,
    ENV_DEBUG_FILE,
    parse_boolean_env,
)

_debug_logger: Optional[logging.Logger] = None
_console: Optional[Console] = None

# Check if we should use Rich console (can be disabled via env var)
_USE_RICH_CONSOLE: bool = parse_boolean_env(
    os.getenv("PLAYBOOKS_DEBUG_USE_RICH", "true")
)


def debug(msg: str = None, **context: Any) -> None:
    """Zero-overhead debug logging when disabled.

    Supports Rich markup syntax for colored output when Rich console is enabled:
        debug("[green]Success![/green] Operation completed")
        debug("Status: [bold red]ERROR[/bold red] - connection failed")

    Args:
        msg: The debug message. Supports Rich markup like [color]text[/color].
             If not provided, will look for 'message' in context.
        **context: Additional context to include in the debug output.
                  Special keys:
                  - 'markup': bool - Enable/disable markup parsing (default: True)
                  - 'color': str - Apply a color to the entire message
                  - 'style': str - Apply a style to the entire message
    """
    if not config.debug:
        return

    # Handle the case where message is passed as keyword argument
    if msg is None and "message" in context:
        msg = context.pop("message")
    elif msg is None:
        msg = "Debug message"  # Default message

    global _debug_logger, _console
    if _debug_logger is None:
        _setup_debug_logger()

    # Extract special context keys
    enable_markup = context.pop("markup", True) if context else True

    # Format the message with context
    if context:
        try:
            # Safely format context, handling circular refs and non-serializable objects
            # Skip color/style/markup keys when formatting output
            context_parts = []
            for k, v in context.items():
                # Skip internal formatting keys
                if k in ["color", "style"]:
                    continue
                try:
                    # Try repr first (safer than str for some objects)
                    context_parts.append(f"{k}={repr(v)}")
                except Exception:
                    # Fallback to type name if repr fails
                    context_parts.append(f"{k}=<{type(v).__name__}>")

            if context_parts:
                context_str = " | ".join(context_parts)
                full_msg = f"{msg} | {context_str}"
            else:
                full_msg = msg
        except Exception:
            # Last resort: just log the message without context
            full_msg = f"{msg} | <context formatting failed>"
    else:
        full_msg = msg

    # Use Rich console for colored output if enabled
    if _USE_RICH_CONSOLE and _console:
        # Also log to file if configured (strip markup for file output)
        if _debug_logger.handlers:
            for handler in _debug_logger.handlers:
                if isinstance(handler, logging.FileHandler):
                    # Strip markup for file logging
                    from rich.text import Text

                    clean_msg = (
                        Text.from_markup(full_msg).plain if enable_markup else full_msg
                    )
                    _debug_logger.debug(clean_msg)
                    break
        _print_with_rich(msg, full_msg, context, enable_markup)
    else:
        # Strip markup for standard logging if markup was in the message
        if enable_markup and "[" in full_msg and "]" in full_msg:
            from rich.text import Text

            full_msg = Text.from_markup(full_msg).plain
        _debug_logger.debug(full_msg)


def _print_with_rich(
    original_msg: str, full_msg: str, context: dict, enable_markup: bool
) -> None:
    """Print debug message using Rich console with colors and markup support.

    Args:
        original_msg: The original message without context (used for auto-color detection)
        full_msg: The formatted message with context
        context: The original context dict for determining colors
        enable_markup: Whether to parse Rich markup in the message
    """
    # If markup is enabled and present in the message, use Rich's markup parsing
    if enable_markup and "[" in full_msg and "]" in full_msg:
        # Print with markup - Rich will handle the colors
        _console.print(
            f"[bold blue]{DEBUG_PREFIX}[/bold blue]{full_msg}", highlight=False
        )
    else:
        # Create styled text based on context (original behavior)
        text = Text()

        # Add prefix with color
        text.append(DEBUG_PREFIX, style="bold blue")

        # Determine color based on context
        style = "white"  # Default color

        # Check if color/style is explicitly specified in context
        if "color" in context:
            style = context.get("color", "white")
        elif "style" in context:
            style = context.get("style", "white")
        else:
            # Auto-detect color based on context keys or message content
            msg_lower = full_msg.lower()

            if (
                any(k in context for k in ["error", "exception", "fail", "failed"])
                or "error" in msg_lower
                or "fail" in msg_lower
                or "exception" in msg_lower
            ):
                style = "bold red"
            elif (
                any(k in context for k in ["warning", "warn"])
                or "warning" in msg_lower
                or "warn" in msg_lower
            ):
                style = "bold yellow"
            elif (
                any(k in context for k in ["success", "complete", "done"])
                or "success" in msg_lower
                or "complete" in msg_lower
            ):
                style = "bold green"
            elif any(k in context for k in ["info", "message"]):
                style = "cyan"
            elif (
                any(k in context for k in ["agent", "agent_id", "process"])
                or "agent" in msg_lower
            ):
                style = "magenta"
            elif (
                any(k in context for k in ["performance", "duration", "time"])
                or "performance" in msg_lower
            ):
                style = "blue"

        # Add the message with determined style
        text.append(full_msg, style=style)

        # Print to console immediately
        _console.print(text, highlight=False)


def _setup_debug_logger() -> None:
    """Setup debug logger once when first debug call is made."""
    global _debug_logger, _console

    _debug_logger = logging.getLogger(DEBUG_LOGGER_NAME)
    _debug_logger.setLevel(logging.DEBUG)

    # Setup Rich console if enabled
    if _USE_RICH_CONSOLE:
        _console = Console(stderr=True)  # Use stderr like logging does

    # Only add standard handler if not using Rich or for file output
    if not _USE_RICH_CONSOLE:
        # Console handler
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(f"{DEBUG_PREFIX}%(message)s"))
        _debug_logger.addHandler(handler)

    # Optional file handler (always add regardless of Rich usage)
    debug_file = os.getenv(ENV_DEBUG_FILE)
    if debug_file:
        file_handler = logging.FileHandler(debug_file)
        file_handler.setFormatter(
            logging.Formatter(f"%(asctime)s {DEBUG_PREFIX}%(message)s")
        )
        _debug_logger.addHandler(file_handler)

    # Prevent propagation to root logger
    _debug_logger.propagate = False


def is_debug_enabled() -> bool:
    """Check if debug logging is enabled."""
    return config.debug



"""Meeting management functionality for AI agents."""

import asyncio
import logging
from datetime import datetime
from typing import Any, Dict, List, Optional

from playbooks.constants import HUMAN_AGENT_KLASS
from playbooks.exceptions import KlassNotFoundError
from playbooks.meetings.meeting_message_handler import MeetingMessageHandler
from playbooks.utils.spec_utils import SpecUtils

from ..agents.base_agent import BaseAgent
from ..debug.debug_handler import debug
from ..llm_messages import MeetingLLMMessage, SessionLogLLMMessage
from ..message import Message, MessageType
from ..playbook import LLMPlaybook, Playbook
from .meeting import JoinedMeeting, Meeting, MeetingInvitation, MeetingInvitationStatus

logger = logging.getLogger(__name__)


class MeetingManager:
    """Manages meeting-related functionality for AI agents."""

    def __init__(self, agent: BaseAgent):
        """Initialize meeting manager.

        Args:
            agent: The agent
        """
        self.agent = agent
        self.meeting_message_handler = MeetingMessageHandler(
            self.agent.id, self.agent.klass
        )

    def ensure_meeting_playbook_kwargs(self, playbooks: Dict[str, Any]) -> None:
        """Ensure that all meeting playbooks have the required kwargs.

        Args:
            playbooks: Dictionary of playbooks to process
        """
        for playbook in playbooks.values():
            if playbook.meeting and isinstance(playbook, LLMPlaybook):
                signature = playbook.signature

                # Check if topic and attendees are missing
                missing_params = []
                if "topic:" not in signature and "topic :" not in signature:
                    missing_params.append("topic: str")
                if "attendees:" not in signature and "attendees :" not in signature:
                    missing_params.append("attendees: List[str]")

                if missing_params:
                    # Find the position to insert parameters (before the closing parenthesis)
                    # Handle cases like "GameRoom() -> None" or "TaxPrepMeeting($form: str) -> None"
                    if ") ->" in signature:
                        # Has return type annotation
                        before_return = signature.split(") ->")[0]
                        after_return = ") ->" + signature.split(") ->", 1)[1]
                    else:
                        # No return type, just ends with )
                        before_return = signature.rstrip(")")
                        after_return = ")"

                    # Check if there are existing parameters
                    if before_return.endswith("("):
                        # No existing parameters, add directly
                        new_params = ", ".join(missing_params)
                    else:
                        # Has existing parameters, add with comma prefix
                        new_params = ", " + ", ".join(missing_params)

                    # Reconstruct the signature
                    playbook.signature = before_return + new_params + after_return

    # def get_meeting_playbooks(self, playbooks: Dict[str, Playbook]) -> List[str]:
    #     """Get list of meeting playbook names.

    #     Args:
    #         playbooks: Dictionary of available playbooks

    #     Returns:
    #         List of playbook names that are marked as meeting playbooks
    #     """
    #     meeting_playbooks = []
    #     for playbook in playbooks.values():
    #         if playbook.meeting:
    #             meeting_playbooks.append(playbook.name)
    #     return meeting_playbooks

    # def is_meeting_playbook(
    #     self, playbook_name: str, playbooks: Dict[str, Playbook]
    # ) -> bool:
    #     """Check if a playbook is a meeting playbook.

    #     Args:
    #         playbook_name: Name of the playbook to check
    #         playbooks: Dictionary of available playbooks

    #     Returns:
    #         True if the playbook is a meeting playbook
    #     """
    #     playbook = playbooks.get(playbook_name)
    #     if not playbook:
    #         return False
    #     return playbook.meeting

    async def create_meeting(
        self, playbook: Playbook, kwargs: Dict[str, Any]
    ) -> Meeting:
        """Create meeting and prepare for invitations.

        Args:
            playbook: The playbook to create a meeting for
            kwargs: Keyword arguments passed to the playbook

        Returns:
            The created meeting
        """
        meeting_id = self.agent.program.meeting_id_registry.generate_meeting_id()

        # Create meeting record
        meeting = Meeting(
            id=meeting_id,
            owner_id=self.agent.id,
            created_at=datetime.now(),
            topic=kwargs.get("topic", f"{playbook} meeting"),
        )

        self.agent.state.owned_meetings[meeting_id] = meeting
        meeting.agent_joined(self.agent)

        # Figure out the attendees
        if "attendees" in kwargs and kwargs["attendees"]:
            # Either provided in the playbook call, e.g. attendees=["agent 1000", "agent 1001", "human"]
            meeting.required_attendees = (
                await self.agent.program.get_agents_by_klasses_or_specs(
                    kwargs["attendees"]
                )
            )

            if not meeting.required_attendees:
                raise ValueError(
                    f"Unknown attendees {kwargs['attendees']} for meeting {meeting_id} with playbook {playbook.name}"
                )

        else:
            # Or configured in playbook metadata, e.g. required_attendees:["Accountant", "user"]
            if playbook.required_attendees or playbook.optional_attendees:
                try:
                    meeting.required_attendees = (
                        await self.agent.program.get_agents_by_klasses(
                            playbook.required_attendees
                        )
                    )

                    meeting.optional_attendees = (
                        await self.agent.program.get_agents_by_klasses(
                            playbook.optional_attendees
                        )
                    )
                except KlassNotFoundError:
                    raise ValueError(
                        f"Unknown required attendees {playbook.required_attendees} or optional attendees {playbook.optional_attendees} for meeting {meeting_id} with playbook {playbook.name}"
                    )

        if not meeting.required_attendees and not meeting.optional_attendees:
            raise ValueError(
                f"Unknown attendees for meeting {meeting_id} with playbook {playbook.name}"
            )

        # Send invitations concurrently
        invitation_tasks = []

        # Create tasks for all required attendees
        for attendee in meeting.required_attendees:
            task = asyncio.create_task(self._invite_to_meeting(meeting, attendee))
            invitation_tasks.append(task)

        # Create tasks for all optional attendees
        for attendee in meeting.optional_attendees:
            task = asyncio.create_task(self._invite_to_meeting(meeting, attendee))
            invitation_tasks.append(task)

        # Yield for other tasks
        await asyncio.sleep(0.1)

        # Wait for all invitations to be sent
        if invitation_tasks:
            await asyncio.gather(*invitation_tasks)

        return meeting

    async def _invite_to_meeting(self, meeting: Meeting, target_agent: BaseAgent):
        """Invite an agent to a meeting.

        Args:
            meeting: The meeting to invite to
            target_agent: The agent to invite
        """
        # Check if the target agent is already a participant
        if meeting.is_participant(target_agent.id):
            response = f"Agent {target_agent.id} is already a participant of meeting {meeting.id}"
        elif target_agent.klass == HUMAN_AGENT_KLASS:
            response = f"User joined meeting {meeting.id}"
            meeting.agent_joined(target_agent)
        else:
            meeting.invitations[target_agent.id] = MeetingInvitation(
                agent=target_agent,
                created_at=datetime.now(),
                status=MeetingInvitationStatus.PENDING,
            )
            response = f"Inviting {str(target_agent)} to meeting {meeting.id}: {meeting.topic or 'Meeting'}"
            self.agent.state.session_log.append(response)
            meeting_msg = MeetingLLMMessage(response, meeting_id=meeting.id)
            self.agent.state.call_stack.add_llm_message(meeting_msg)
            asyncio.create_task(self._send_invitation(meeting, target_agent))

        return response

    async def _send_invitation(self, meeting: Meeting, target_agent: BaseAgent):
        """Send meeting invitation to an agent using the message system.

        Args:
            meeting: The meeting to send the invitation to
            target_agent: The agent to send the invitation to
        """

        # Send structured invitation message
        invitation_content = meeting.topic or "Meeting"

        await self.agent.program.route_message(
            sender_id=self.agent.id,
            sender_klass=self.agent.klass,
            receiver_spec=SpecUtils.to_agent_spec(target_agent.id),
            message=invitation_content,
            message_type=MessageType.MEETING_INVITATION,
            meeting_id=meeting.id,
        )

        response = (
            f"Invited {str(target_agent)} to meeting {meeting.id}: {invitation_content}"
        )
        self.agent.state.session_log.append(response)
        meeting_msg = MeetingLLMMessage(response, meeting_id=meeting.id)
        self.agent.state.call_stack.add_llm_message(meeting_msg)

        return response

    async def InviteToMeeting(self, meeting_id: str, attendees: List[str]):
        """Invite an agent to a meeting.

        Args:
            meeting_id: The ID of the meeting to invite to
            attendees: The agents to invite
        """
        if meeting_id not in self.agent.state.owned_meetings:
            return f"I am not the owner of meeting {meeting_id}, so cannot invite attendees"

        meeting = self.agent.state.owned_meetings[meeting_id]
        attendees = self.agent.program.get_agents_by_specs(attendees)

        # Send invitations concurrently
        invitation_tasks = []
        for attendee in attendees:
            task = asyncio.create_task(self._invite_to_meeting(meeting, attendee))
            invitation_tasks.append(task)

        # Wait for all invitations and collect responses
        responses = await asyncio.gather(*invitation_tasks) if invitation_tasks else []
        return "\n".join(responses)

    async def _wait_for_required_attendees(
        self, meeting: Meeting, timeout_seconds: int = 30
    ):
        """Wait for required attendees to join the meeting before proceeding.

        Args:
            meeting: The meeting to wait for
            timeout_seconds: Maximum time to wait for attendees

        Raises:
            TimeoutError: If required attendees don't join within timeout
            ValueError: If required attendee rejects the invitation
        """
        # If no attendees to wait for, proceed immediately
        if not meeting.required_attendees:
            message = f"No required attendees to wait for in meeting {meeting.id} - proceeding immediately"
            self.agent.state.session_log.append(message)
            meeting_msg = MeetingLLMMessage(message, meeting_id=meeting.id)
            self.agent.state.call_stack.add_llm_message(meeting_msg)
            return

        messages = f"Waiting for required attendees to join meeting {meeting.id}: {[attendee.__repr__() for attendee in meeting.required_attendees]}"
        self.agent.state.session_log.append(messages)
        meeting_msg = MeetingLLMMessage(messages, meeting_id=meeting.id)
        self.agent.state.call_stack.add_llm_message(meeting_msg)

        # Track which required attendees have joined
        start_time = asyncio.get_event_loop().time()

        while meeting.has_pending_invitations():
            # Check for timeout
            if asyncio.get_event_loop().time() - start_time > timeout_seconds:
                raise TimeoutError(
                    f"Timeout waiting for required attendees to join meeting {meeting.id}. "
                    f"Missing: {[attendee.__repr__() for attendee in meeting.missing_required_attendees()]}"
                )
            message = f"Waiting for required attendees to join meeting {meeting.id}: {[attendee.__repr__() for attendee in meeting.missing_required_attendees()]}"
            self.agent.state.session_log.append(message)
            await asyncio.sleep(1)

        message = f"All required attendees have joined meeting {meeting.id}: {[attendee.__repr__() for attendee in meeting.joined_attendees]}"
        self.agent.state.session_log.append(message)
        meeting_msg = MeetingLLMMessage(message, meeting_id=meeting.id)
        self.agent.state.call_stack.add_llm_message(meeting_msg)

        # Finally, set the meeting ID as the current meeting ID in the call stack
        self.agent.state.call_stack.peek().meeting_id = meeting.id

    def get_current_meeting_from_call_stack(self) -> Optional[str]:
        """Get meeting ID from top meeting playbook in call stack.

        Args:
            call_stack: The agent's call stack

        Returns:
            Meeting ID if currently in a meeting, None otherwise
        """
        call_stack = self.agent.state.call_stack
        for frame in reversed(call_stack.frames):
            if frame.is_meeting and frame.meeting_id:
                return frame.meeting_id
        return None

    async def broadcast_to_meeting_as_owner(
        self,
        meeting_id: str,
        message: str,
        from_agent_id: str = None,
        from_agent_klass: str = None,
    ):
        """Broadcast a message to all participants in a meeting.

        Args:
            meeting_id: ID of the meeting to broadcast to
            message: Message content to send
        """
        # Check if I'm the owner of this meeting
        assert (
            hasattr(self.agent, "state")
            and hasattr(self.agent.state, "owned_meetings")
            and meeting_id in self.agent.state.owned_meetings
        )

        if not from_agent_id or not from_agent_klass:
            from_agent_id = self.agent.id
            from_agent_klass = self.agent.klass

        meeting: Meeting = self.agent.state.owned_meetings[meeting_id]
        self.agent.state.session_log.append(
            f"Adding message to owned meeting {meeting_id}: {message}"
        )
        for participant in meeting.joined_attendees:
            if participant.id == from_agent_id:
                continue
            await self.agent.program.route_message(
                sender_id=from_agent_id,
                sender_klass=from_agent_klass,
                receiver_spec=SpecUtils.to_agent_spec(participant.id),
                message=message,
                message_type=MessageType.MEETING_BROADCAST,
                meeting_id=meeting_id,
            )

    async def broadcast_to_meeting_as_participant(self, meeting_id: str, message: str):
        """Broadcast a message to all participants in a meeting.

        Args:
            meeting_id: ID of the meeting to broadcast to
            message: Message content to send
        """
        assert (
            hasattr(self.agent, "state")
            and hasattr(self.agent.state, "joined_meetings")
            and meeting_id in self.agent.state.joined_meetings
        )
        # I'm a participant - I will request the owner
        owner_id = self.agent.state.joined_meetings[meeting_id].owner_id

        # Send to meeting owner with protocol prefix
        await self.agent.program.route_message(
            sender_id=self.agent.id,
            sender_klass=self.agent.klass,
            receiver_spec=SpecUtils.to_agent_spec(owner_id),
            message=message,
            message_type=MessageType.MEETING_BROADCAST_REQUEST,
            meeting_id=meeting_id,
        )

    async def _add_message_to_buffer(self, message: Message) -> bool:
        """Add a message to buffer and notify waiting processes.

        This is the single entry point for all incoming messages.
        """
        if message.message_type == MessageType.MEETING_INVITATION:
            # Process meeting invitation immediately
            return await self._handle_meeting_invitation_immediately(message)
        elif message.message_type == MessageType.MEETING_INVITATION_RESPONSE:
            # Process meeting response immediately
            await self._handle_meeting_response_immediately(message)
            return True
        elif message.message_type == MessageType.MEETING_BROADCAST_REQUEST:
            # Process meeting message immediately
            await self.broadcast_to_meeting_as_owner(
                meeting_id=message.meeting_id,
                message=message.content,
                from_agent_id=message.sender_id,
                from_agent_klass=message.sender_klass,
            )
            return True
        return False

    async def _handle_meeting_invitation_immediately(self, message) -> None:
        """Handle meeting invitation immediately without buffering."""
        # Extract meeting information from the message
        meeting_id = getattr(message, "meeting_id", None)
        sender_id = message.sender_id
        topic = message.content  # The invitation content contains the topic

        if meeting_id:
            # Use async task to handle the invitation since this is called synchronously
            return await self._process_meeting_invitation(sender_id, meeting_id, topic)
        else:
            log = f"Received meeting invitation from {sender_id} without meeting_id"
            self.agent.state.session_log.append(log)
            session_msg = SessionLogLLMMessage(log, log_level="warning")
            self.agent.state.call_stack.add_llm_message(session_msg)
            return True

    async def _handle_meeting_response_immediately(self, message) -> None:
        """Handle meeting response immediately without buffering."""
        # Use existing meeting message handler for responses
        if hasattr(self, "meeting_message_handler"):
            # Create task to handle response asynchronously
            await self.meeting_message_handler.handle_meeting_response(
                message,
                self.agent,
            )
        else:
            self.agent.state.session_log.append(
                f"Received meeting response from {message.sender_id} but no handler available"
            )

    async def _process_meeting_invitation(
        self, inviter_id: str, meeting_id: str, topic: str
    ):
        """Process a meeting invitation by checking for suitable meeting playbooks.

        Args:
            inviter_id: ID of the agent that sent the invitation
            meeting_id: ID of the meeting
            topic: Topic/description of the meeting
        """
        log = f"Received meeting invitation for meeting {meeting_id} from {inviter_id} for '{topic}'"
        self.agent.state.session_log.append(log)
        meeting_msg = MeetingLLMMessage(log, meeting_id=meeting_id)
        self.agent.state.call_stack.add_llm_message(meeting_msg)

        # Check if agent is busy (has active call stack)
        if (
            "$_busy" in self.agent.state.variables
            and self.agent.state.variables["$_busy"].value == "true"
        ):
            log = f"Rejecting meeting {meeting_id} - agent is busy"
            self.agent.state.session_log.append(log)
            meeting_msg = MeetingLLMMessage(log, meeting_id=meeting_id)
            self.agent.state.call_stack.add_llm_message(meeting_msg)
            if self.agent.program:
                await self.agent.program.route_message(
                    sender_id=self.agent.id,
                    sender_klass=self.agent.klass,
                    receiver_spec=SpecUtils.to_agent_spec(inviter_id),
                    message=f"REJECTED {meeting_id}",
                    message_type=MessageType.MEETING_INVITATION_RESPONSE,
                    meeting_id=meeting_id,
                )
            return True
        return False

        # # Find matching meeting playbooks
        # meeting_playbooks = []
        # for playbook in self.agent.playbooks.values():
        #     if playbook.meeting:
        #         meeting_playbooks.append(playbook)

        # if not meeting_playbooks:
        #     # No meeting playbooks available
        #     log = f"Rejecting meeting {meeting_id} - no meeting playbooks available"
        #     self.agent.state.session_log.append(log)
        #     meeting_msg = MeetingLLMMessage(log, meeting_id=meeting_id)
        #     self.agent.state.call_stack.add_llm_message(meeting_msg)
        #     available_types = self.get_meeting_playbooks()
        #     rejection_message = f"Meeting invitation rejected: Cannot handle this type of meeting. Available meeting types: {available_types}"

        #     if self.agent.program:
        #         await self.agent.program.route_message(
        #             self.agent.id,
        #             inviter_id,
        #             f"REJECTED {meeting_id}: {rejection_message}",
        #             message_type="meeting_response",
        #             meeting_id=meeting_id,
        #         )
        #     return

    async def _accept_meeting_invitation(
        self, meeting_id: str, inviter_id: str, topic: str, playbook_name: str
    ) -> bool:
        # Accept the invitation and join the meeting
        debug(f"{str(self.agent)}: Accepting meeting invitation {meeting_id}")
        log = f"Accepting meeting invitation {meeting_id}"
        meeting_id = SpecUtils.extract_meeting_id(meeting_id)
        self.agent.state.session_log.append(log)
        meeting_msg = MeetingLLMMessage(log, meeting_id=meeting_id)
        self.agent.state.call_stack.add_llm_message(meeting_msg)

        # Store meeting info in joined_meetings for future message routing

        self.agent.state.joined_meetings[meeting_id] = JoinedMeeting(
            id=meeting_id,
            owner_id=inviter_id,
            joined_at=datetime.now(),
        )
        debug(f"{str(self.agent)}: joined_meetings {self.agent.state.joined_meetings}")

        # Send structured JOINED response
        if self.agent.program:
            await self.agent.program.route_message(
                sender_id=self.agent.id,
                sender_klass=self.agent.klass,
                receiver_spec=SpecUtils.to_agent_spec(inviter_id),
                message=f"JOINED {meeting_id}",
                message_type=MessageType.MEETING_INVITATION_RESPONSE,
                meeting_id=meeting_id,
            )

        # The initiator will add us as a participant when they receive our JOINED message
        # We don't directly access the meeting object here to support remote agents

        # Execute the first available meeting playbook
        # In a more sophisticated implementation, we could match playbook by topic/type
        meeting_playbook = self.agent.playbooks[playbook_name]

        try:
            log = f"Starting meeting playbook '{meeting_playbook.name}' for meeting {meeting_id}"
            debug(f"{str(self.agent)}: {log}")
            self.agent.state.session_log.append(log)
            meeting_msg = MeetingLLMMessage(log, meeting_id=meeting_id)
            self.agent.state.call_stack.add_llm_message(meeting_msg)

            # Execute the meeting playbook with meeting context
            debug(
                "Agent executing meeting playbook",
                agent_id=self.agent.id,
                playbook_name=meeting_playbook.name,
            )
            task = asyncio.create_task(
                self.agent.execute_playbook(
                    meeting_playbook.name,
                    args=[],
                    kwargs={"meeting_id": meeting_id, "topic": topic},
                )
            )
            task.add_done_callback(
                lambda t: debug(
                    "Meeting playbook task done",
                    agent=str(self.agent),
                    playbook_name=meeting_playbook.name,
                )
            )
            await asyncio.gather(task)

        except Exception as e:
            log = f"Error executing meeting playbook for {meeting_id}: {str(e)}"
            self.agent.state.session_log.append(log)
            meeting_msg = MeetingLLMMessage(log, meeting_id=meeting_id)
            self.agent.state.call_stack.add_llm_message(meeting_msg)
            # Send error message to meeting
            if self.agent.program:
                await self.agent.program.route_message(
                    sender_id=self.agent.id,
                    sender_klass=self.agent.klass,
                    receiver_spec=SpecUtils.to_agent_spec(inviter_id),
                    message=f"Meeting {meeting_id}: Error in playbook execution - {str(e)}",
                    message_type=MessageType.MEETING_INVITATION_RESPONSE,
                    meeting_id=meeting_id,
                )

    # async def initialize_meeting_playbook(
    #     self,
    #     playbook_name: str,
    #     kwargs: Dict[str, Any],
    #     playbooks: Dict[str, Any],
    #     meeting_registry: MeetingRegistry,
    #     session_log,
    #     wait_for_attendees_callback,
    # ):
    #     """Initialize meeting before executing meeting playbook.

    #     This method is called implicitly before any meeting playbook executes.
    #     For new meetings, it creates the meeting, sends invitations, and waits for required participants.
    #     For existing meetings (when meeting_id is provided), it joins the existing meeting.

    #     Args:
    #         playbook_name: Name of the meeting playbook being executed
    #         kwargs: Keyword arguments passed to the playbook
    #         playbooks: Dictionary of available playbooks
    #         meeting_registry: Registry for meeting IDs
    #         session_log: Session log for recording events
    #         wait_for_attendees_callback: Callback to wait for required attendees
    #     """
    #     # Check if we're joining an existing meeting (meeting_id provided) or creating a new one
    #     existing_meeting_id = kwargs.get("meeting_id")

    #     if existing_meeting_id:
    #         # Joining an existing meeting - just proceed with execution
    #         session_log.append(
    #             f"Joining existing meeting {existing_meeting_id} for playbook {playbook_name}"
    #         )
    #         return  # No need to create meeting or wait for attendees

    #     # Creating a new meeting
    #     kwargs_attendees = kwargs.get("attendees", [])
    #     topic = kwargs.get("topic", f"{playbook_name} meeting")

    #     # Determine attendee strategy: kwargs attendees take precedence
    #     if kwargs_attendees:
    #         # If attendees specified in kwargs, treat them as required
    #         required_attendees = kwargs_attendees
    #         all_attendees = kwargs_attendees
    #         session_log.append(
    #             f"Using kwargs attendees as required for meeting {playbook_name}: {required_attendees}"
    #         )
    #     else:
    #         # If no kwargs attendees, use metadata-defined attendees
    #         metadata_required, metadata_optional = self.get_playbook_attendees(
    #             playbook_name, playbooks
    #         )
    #         required_attendees = metadata_required
    #         all_attendees = list(set(metadata_required + metadata_optional))
    #         session_log.append(
    #             f"Using metadata attendees for meeting {playbook_name}: required={metadata_required}, optional={metadata_optional}"
    #         )

    #     # Filter out the requester from required attendees (they're already present)
    #     required_attendees_to_wait_for = [
    #         attendee
    #         for attendee in required_attendees
    #         if attendee != self.agent_klass and attendee != self.agent_id
    #     ]

    #     # Create the meeting
    #     meeting_id = await self.create_meeting(
    #         invited_agents=all_attendees, topic=topic, meeting_registry=meeting_registry
    #     )

    #     # Store meeting_id in kwargs for the playbook to access
    #     kwargs["meeting_id"] = meeting_id

    #     # Log the meeting initialization
    #     session_log.append(
    #         f"Initialized meeting {meeting_id} for playbook {playbook_name}"
    #     )

    #     # Wait for required attendees to join before proceeding
    #     await wait_for_attendees_callback(meeting_id, required_attendees_to_wait_for)



import enum
import uuid
from dataclasses import dataclass
from datetime import datetime
from typing import Optional


class MessageType(enum.Enum):
    DIRECT = "direct"
    MEETING_BROADCAST_REQUEST = "meeting_broadcast_request"
    MEETING_BROADCAST = "meeting_broadcast"
    MEETING_INVITATION = "meeting_invitation"
    MEETING_INVITATION_RESPONSE = "meeting_invitation_response"


@dataclass
class Message:
    """Represents a message in the system."""

    sender_id: str
    sender_klass: str

    recipient_id: Optional[str]
    recipient_klass: Optional[str]

    message_type: MessageType
    content: str

    meeting_id: Optional[str]

    id: str = uuid.uuid4()
    created_at: datetime = datetime.now()

    def __str__(self):
        meeting_message = f", in meeting {self.meeting_id}" if self.meeting_id else ""
        message_type = (
            "[MEETING INVITATION] "
            if self.message_type == MessageType.MEETING_INVITATION
            else ""
        )
        return f"{message_type}Message from {self.sender_klass}(agent {self.sender_id}) to {self.recipient_klass}(agent {self.recipient_id}){meeting_message}: {self.content}"



from enum import Enum


class AgentType(str, Enum):
    HUMAN = "human"
    AI = "ai"


class RoutingType(str, Enum):
    DIRECT = "direct"
    BROADCAST = "broadcast"


class LLMMessageRole(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"


class LLMMessageType(str, Enum):
    # Core execution types
    SYSTEM_PROMPT = "system_prompt"
    USER_INPUT = "user_input"
    ASSISTANT_RESPONSE = "assistant_response"

    # Playbook execution types
    PLAYBOOK_IMPLEMENTATION = "playbook_implementation"
    EXECUTION_RESULT = "execution_result"
    TRIGGER_INSTRUCTIONS = "trigger_instructions"

    # Communication types
    AGENT_COMMUNICATION = "agent_communication"
    MEETING_MESSAGE = "meeting_message"

    # Agent information types
    AGENT_INFO = "agent_info"
    OTHER_AGENT_INFO = "other_agent_info"

    # Data types
    FILE_LOAD = "file_load"
    SESSION_LOG = "session_log"


class StartupMode(str, Enum):
    DEFAULT = "default"
    STANDBY = "standby"


class LLMExecutionMode(str, Enum):
    """Execution modes for LLM playbooks."""

    PLAYBOOK = "playbook"  # Traditional structured steps (default)
    REACT = "react"  # Loops with tool calls until exit conditions
    RAW = "raw"  # One LLM call, no loops or structure



import logging
from typing import Any, Callable, Dict, Optional

from .base import Playbook

logger = logging.getLogger(__name__)


class RemotePlaybook(Playbook):
    """A playbook that executes on a remote system.

    This class represents playbooks that are executed remotely, such as
    through MCP (Model Context Protocol) or other remote execution mechanisms.
    """

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        agent_name: Optional[str] = None,
        execute_fn: Optional[Callable] = None,
        parameters: Optional[Dict[str, Any]] = None,
        timeout: Optional[float] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ):
        """Initialize a remote playbook.

        Args:
            name: The name of the playbook
            description: Human-readable description of the playbook
            agent_name: Name of the agent this playbook belongs to
            execute_fn: Function to call for remote execution
            parameters: Schema describing the expected parameters
            timeout: Timeout for remote execution in seconds
            metadata: Additional metadata for the playbook
        """
        super().__init__(name, description, agent_name, metadata)
        self.execute_fn = execute_fn
        self.parameters = parameters or {}
        self.timeout = timeout

    @property
    def signature(self) -> str:
        """Get the signature of the remote playbook.

        Returns:
            The signature of the remote playbook
        """
        return f"{self.name}({', '.join([f'{param}: {type(param).__name__}' for param in self.parameters])})"

    async def execute(self, *args, **kwargs) -> Any:
        """Execute the remote playbook.

        Args:
            *args: Positional arguments for the playbook
            **kwargs: Keyword arguments for the playbook

        Returns:
            The result of executing the remote playbook

        Raises:
            ValueError: If no execution function is configured
            Exception: If remote execution fails
        """
        if not self.execute_fn:
            raise ValueError(f"Remote playbook {self.name} has no execution function")

        logger.debug(
            f"Executing remote playbook {self.name} with args={args}, kwargs={kwargs}"
        )

        try:
            # Execute the remote function
            if self.timeout:
                # TODO: Add timeout handling when we implement the transport layer
                result = await self.execute_fn(*args, **kwargs)
            else:
                result = await self.execute_fn(*args, **kwargs)

            logger.debug(f"Remote playbook {self.name} completed successfully")
            return result
        except Exception as e:
            logger.error(f"Remote playbook {self.name} failed: {str(e)}")
            raise

    def get_parameters(self) -> Dict[str, Any]:
        """Get the parameters schema for this playbook.

        Returns:
            A dictionary describing the expected parameters
        """
        return self.parameters

    def get_description(self) -> str:
        """Get a human-readable description of this playbook.

        Returns:
            The description if available, otherwise the name
        """
        return self.description or self.name

    def create_namespace_function(self, agent) -> "Callable":
        """RemotePlaybooks should not be added to namespace.

        Raises:
            NotImplementedError: RemotePlaybooks don't support namespace functions
        """
        raise NotImplementedError("RemotePlaybooks should not be added to namespace")

    def __repr__(self) -> str:
        """Return a string representation of the playbook."""
        return f"RemotePlaybook {self.agent_name}.{self.name}: {self.description}"



"""Clean semantic LLM message types with minimal, maintainable design."""

from typing import Any, Dict, Optional

from playbooks.enums import LLMMessageRole, LLMMessageType

from .base import LLMMessage

# Core semantic message types - minimal set covering all use cases


class SystemPromptLLMMessage(LLMMessage):
    """System prompts and instructions - cached by default."""

    def __init__(self, content: str) -> None:
        super().__init__(
            content=content,
            role=LLMMessageRole.SYSTEM,
            type=LLMMessageType.SYSTEM_PROMPT,
            cached=True,
        )


class UserInputLLMMessage(LLMMessage):
    """User inputs and instructions - not cached by default."""

    def __init__(self, content: str, cached: bool = False) -> None:
        super().__init__(
            content=content,
            role=LLMMessageRole.USER,
            type=LLMMessageType.USER_INPUT,
            cached=cached,
        )

    def to_compact_message(self) -> Optional[Dict[str, Any]]:
        """Remove user inputs during compaction."""
        return None


class AssistantResponseLLMMessage(LLMMessage):
    """LLM responses - cached by default for conversation context."""

    def __init__(self, content: str, cached: bool = True) -> None:
        super().__init__(
            content=content,
            role=LLMMessageRole.ASSISTANT,
            type=LLMMessageType.ASSISTANT_RESPONSE,
            cached=cached,
        )

    def to_compact_message(self) -> Dict[str, Any]:
        """Use first line that begins with 'recap -' for compaction."""
        lines = self.content.split("\n")

        # We don't want to confuse LLM with a partial assistant response
        # so we use the user role for the compacted message
        compact_message_role = LLMMessageRole.USER

        # Find line that begins with 'recap -'
        recap_line = next(
            (line.strip() for line in lines if line.strip().startswith("recap -")), None
        )
        if recap_line:
            return {
                "role": compact_message_role,
                "content": recap_line.replace("recap -", "").strip(),
            }

        # If no recap line found, use first line
        first_line = lines[0].strip() if lines else ""
        return {"role": compact_message_role, "content": first_line}


class PlaybookImplementationLLMMessage(LLMMessage):
    """Playbook markdown implementation - cached for performance."""

    def __init__(self, content: str, playbook_name: str) -> None:
        self.playbook_name = self._validate_string_param(playbook_name, "playbook_name")

        super().__init__(
            content=content,
            role=LLMMessageRole.USER,
            type=LLMMessageType.PLAYBOOK_IMPLEMENTATION,
            cached=True,
        )

    def __eq__(self, other: object) -> bool:
        """Check equality including custom attributes."""
        if not isinstance(other, self.__class__):
            return False
        return super().__eq__(other) and self.playbook_name == other.playbook_name

    def __hash__(self) -> int:
        """Hash including custom attributes."""
        return hash(
            (
                self.__class__.__name__,
                self.content,
                self.role,
                self.type,
                self.cached,
                self.playbook_name,
            )
        )


class ExecutionResultLLMMessage(LLMMessage):
    """Playbook execution results - not cached as they're context-specific."""

    def __init__(self, content: str, playbook_name: str, success: bool = True) -> None:
        self.playbook_name = self._validate_string_param(playbook_name, "playbook_name")
        if not isinstance(success, bool):
            raise TypeError(f"success must be a boolean, got {type(success).__name__}")
        self.success = success

        super().__init__(
            content=content,
            role=LLMMessageRole.USER,
            type=LLMMessageType.EXECUTION_RESULT,
            cached=False,
        )

    def __eq__(self, other: object) -> bool:
        """Check equality including custom attributes."""
        if not isinstance(other, self.__class__):
            return False
        return (
            super().__eq__(other)
            and self.playbook_name == other.playbook_name
            and self.success == other.success
        )

    def __hash__(self) -> int:
        """Hash including custom attributes."""
        return hash(
            (
                self.__class__.__name__,
                self.content,
                self.role,
                self.type,
                self.cached,
                self.playbook_name,
                self.success,
            )
        )


class AgentCommunicationLLMMessage(LLMMessage):
    """Inter-agent communications - not cached as they're event-specific."""

    def __init__(self, content: str, sender_agent: str, target_agent: str) -> None:
        self.sender_agent = self._validate_string_param(sender_agent, "sender_agent")
        self.target_agent = self._validate_string_param(target_agent, "target_agent")

        # Note: sender_agent can be the same as target_agent in meeting contexts
        # or when an agent is processing its own messages

        super().__init__(
            content=content,
            role=LLMMessageRole.USER,
            type=LLMMessageType.AGENT_COMMUNICATION,
            cached=False,
        )

    def __eq__(self, other: object) -> bool:
        """Check equality including custom attributes."""
        if not isinstance(other, self.__class__):
            return False
        return (
            super().__eq__(other)
            and self.sender_agent == other.sender_agent
            and self.target_agent == other.target_agent
        )

    def __hash__(self) -> int:
        """Hash including custom attributes."""
        return hash(
            (
                self.__class__.__name__,
                self.content,
                self.role,
                self.type,
                self.cached,
                self.sender_agent,
                self.target_agent,
            )
        )


class MeetingLLMMessage(LLMMessage):
    """Meeting-related communications - not cached as they're event-specific."""

    def __init__(self, content: str, meeting_id: str) -> None:
        self.meeting_id = self._validate_string_param(meeting_id, "meeting_id")

        super().__init__(
            content=content,
            role=LLMMessageRole.USER,
            type=LLMMessageType.MEETING_MESSAGE,
            cached=False,
        )

    def __eq__(self, other: object) -> bool:
        """Check equality including custom attributes."""
        if not isinstance(other, self.__class__):
            return False
        return super().__eq__(other) and self.meeting_id == other.meeting_id

    def __hash__(self) -> int:
        """Hash including custom attributes."""
        return hash(
            (
                self.__class__.__name__,
                self.content,
                self.role,
                self.type,
                self.cached,
                self.meeting_id,
            )
        )


class TriggerInstructionsLLMMessage(LLMMessage):
    """Playbook trigger instructions - cached for performance."""

    def __init__(self, content: str) -> None:
        super().__init__(
            content=content,
            role=LLMMessageRole.USER,
            type=LLMMessageType.TRIGGER_INSTRUCTIONS,
            cached=True,
        )


class AgentInfoLLMMessage(LLMMessage):
    """Current agent information - cached as it doesn't change often."""

    def __init__(self, content: str) -> None:
        super().__init__(
            content=content,
            role=LLMMessageRole.USER,
            type=LLMMessageType.AGENT_INFO,
            cached=True,
        )


class OtherAgentInfoLLMMessage(LLMMessage):
    """Other available agents information - cached as it doesn't change often."""

    def __init__(self, content: str) -> None:
        super().__init__(
            content=content,
            role=LLMMessageRole.USER,
            type=LLMMessageType.OTHER_AGENT_INFO,
            cached=True,
        )


class FileLoadLLMMessage(LLMMessage):
    """File content loading - not cached as files may change."""

    def __init__(self, content: str, file_path: str) -> None:
        self.file_path = self._validate_string_param(file_path, "file_path")
        # Note: Content size validation is now handled by base class

        super().__init__(
            content=content,
            role=LLMMessageRole.USER,
            type=LLMMessageType.FILE_LOAD,
            cached=False,
        )

    def __eq__(self, other: object) -> bool:
        """Check equality including custom attributes."""
        if not isinstance(other, self.__class__):
            return False
        return super().__eq__(other) and self.file_path == other.file_path

    def __hash__(self) -> int:
        """Hash including custom attributes."""
        return hash(
            (
                self.__class__.__name__,
                self.content,
                self.role,
                self.type,
                self.cached,
                self.file_path,
            )
        )


class SessionLogLLMMessage(LLMMessage):
    """Session logging and status updates - not cached."""

    def __init__(self, content: str, log_level: str = "INFO") -> None:
        # Validate log level
        valid_levels = {"DEBUG", "INFO", "WARN", "ERROR", "FATAL"}
        if not isinstance(log_level, str):
            raise TypeError(
                f"log_level must be a string, got {type(log_level).__name__}"
            )
        if log_level not in valid_levels:
            raise ValueError(
                f"log_level must be one of {valid_levels}, got {log_level!r}"
            )
        self.log_level = log_level

        super().__init__(
            content=content,
            role=LLMMessageRole.SYSTEM,  # Fixed: logs are system-level information
            type=LLMMessageType.SESSION_LOG,
            cached=False,
        )

    def __eq__(self, other: object) -> bool:
        """Check equality including custom attributes."""
        if not isinstance(other, self.__class__):
            return False
        return super().__eq__(other) and self.log_level == other.log_level

    def __hash__(self) -> int:
        """Hash including custom attributes."""
        return hash(
            (
                self.__class__.__name__,
                self.content,
                self.role,
                self.type,
                self.cached,
                self.log_level,
            )
        )



from abc import abstractmethod
from typing import Any, Dict, Optional

from rich.console import Console

from playbooks.debug_logger import debug
from playbooks.exceptions import ExecutionFinished

from .base import Playbook


class LocalPlaybook(Playbook):
    """Abstract base class for playbooks that execute locally.

    This class provides common functionality for playbooks that run in the local
    environment, including error handling and logging.
    """

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        agent_name: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        source_file_path: Optional[str] = None,
        source_line_number: Optional[int] = None,
    ):
        """Initialize a local playbook.

        Args:
            name: The name/class of the playbook
            description: Human-readable description of the playbook
            agent_name: Name of the agent this playbook belongs to
            source_file_path: The file path of the source where this playbook is defined
            source_line_number: The line number in the source where this playbook is defined
        """
        super().__init__(
            name=name,
            description=description,
            agent_name=agent_name,
            metadata=metadata,
        )
        self.source_file_path = source_file_path
        self.source_line_number = source_line_number

    async def execute(self, *args, **kwargs) -> Any:
        """Execute the local playbook with error handling and logging.

        Args:
            *args: Positional arguments for the playbook
            **kwargs: Keyword arguments for the playbook

        Returns:
            The result of executing the playbook

        Raises:
            Exception: If execution fails
        """
        debug(f"Executing local playbook {self.name} with args={args}, kwargs={kwargs}")

        try:
            result = await self._execute_impl(*args, **kwargs)
            debug(f"Local playbook {self.name} completed successfully")
            return result
        except ExecutionFinished:
            raise
        except Exception as e:
            error_msg = f"Local playbook {self.name} failed: {str(e)}"
            Console(stderr=True).print(f"[bold red]ERROR:[/bold red] {error_msg}")
            return f"Error: {str(e)}"

    @abstractmethod
    async def _execute_impl(self, *args, **kwargs) -> Any:
        """Implementation-specific execution logic.

        Subclasses must implement this method to define their execution behavior.

        Args:
            *args: Positional arguments for the playbook
            **kwargs: Keyword arguments for the playbook

        Returns:
            The result of executing the playbook
        """
        pass

    def get_description(self) -> str:
        """Get a human-readable description of this playbook.

        Returns:
            The description if available, otherwise the name
        """
        return self.description or self.name



"""Utility functions for counting tokens using tiktoken."""

import json
from typing import Dict, List, Union

import tiktoken


def get_token_count(text: str, model: str = "gpt-4") -> int:
    """
    Count the number of tokens in a text string using tiktoken.

    Args:
        text: The text to count tokens for
        model: The model to use for tokenization (default: gpt-4)

    Returns:
        The number of tokens in the text
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
        return len(encoding.encode(text))
    except KeyError:
        # Fallback to cl100k_base encoding if model not found
        encoding = tiktoken.get_encoding("cl100k_base")
        return len(encoding.encode(text))


def get_messages_token_count(
    messages: List[Dict[str, str]], model: str = "gpt-4"
) -> int:
    """
    Count the total number of tokens in a list of messages.

    Args:
        messages: List of message dictionaries with 'role' and 'content' keys
        model: The model to use for tokenization (default: gpt-4)

    Returns:
        The total number of tokens in all messages
    """
    total_tokens = 0

    for message in messages:
        # Count tokens for the message content
        content = message.get("content", "")
        total_tokens += get_token_count(content, model)

        # Add tokens for message structure (role, etc.)
        # OpenAI's ChatML format adds some overhead per message
        total_tokens += 4  # Approximate overhead per message

        # Add tokens for role
        role = message.get("role", "")
        total_tokens += get_token_count(role, model)

    # Add tokens for conversation structure
    total_tokens += 2  # Approximate overhead for the conversation

    return total_tokens


def get_dict_token_count(data: Union[Dict, List], model: str = "gpt-4") -> int:
    """
    Count the number of tokens in a dictionary or list by converting to JSON.

    Args:
        data: The dictionary or list to count tokens for
        model: The model to use for tokenization (default: gpt-4)

    Returns:
        The number of tokens in the JSON representation
    """
    json_str = json.dumps(data, indent=2)
    return get_token_count(json_str, model)



class AsyncInitMixin:
    _ALLOW_INIT = False

    def __init__(self, *args, **kwargs):
        if not self.__class__._ALLOW_INIT:
            raise RuntimeError(
                f"{self.__class__.__name__} cannot be instantiated directly. "
                f"Use 'await {self.__class__.__name__}.create()' instead."
            )
        super().__init__(*args, **kwargs)

    @classmethod
    async def create(cls, *args, **kwargs):
        cls._ALLOW_INIT = True
        try:
            instance = cls(*args, **kwargs)
        finally:
            cls._ALLOW_INIT = False
        await instance._async_init()
        return instance

    async def _async_init(self):
        """Override this method in subclasses for async initialization"""
        pass



"""LLM Context Compaction for managing growing message context in playbooks."""

import os
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from playbooks.llm_messages import (
    AssistantResponseLLMMessage,
    LLMMessage,
    UserInputLLMMessage,
)


@dataclass
class CompactionConfig:
    """Configuration for LLM context compaction."""

    # Core strategy parameters
    min_preserved_assistant_messages: int = (
        None  # Always keep last N assistant messages
    )
    batch_size: int = None  # Compact in batches of N

    # Feature toggles
    enabled: bool = None  # Master enable/disable

    def __post_init__(self):
        """Initialize from environment variables if values not provided."""
        if self.enabled is None:
            self.enabled = os.getenv("LLM_COMPACTION_ENABLED", "true").lower() == "true"

        if self.min_preserved_assistant_messages is None:
            self.min_preserved_assistant_messages = int(
                os.getenv("LLM_COMPACTION_MIN_PRESERVED_ASSISTANT_MESSAGES", "1")
            )

        if self.batch_size is None:
            self.batch_size = int(os.getenv("LLM_COMPACTION_BATCH_SIZE", "3"))


class LLMContextCompactor:
    """Handles progressive compaction of LLM messages."""

    def __init__(self, config: Optional[CompactionConfig] = None):
        self.config = config or CompactionConfig()

    def compact_messages(self, messages: List[LLMMessage]) -> List[Dict[str, Any]]:
        """Main entry point for message compaction - returns LLM API format."""
        if not self.config.enabled or len(messages) == 0:
            return [msg.to_full_message() for msg in messages]

        # Walk backwards to find last safe (uncompacted) assistant message
        assistant_count = 0
        safe_assistant_index = -1
        for i in range(len(messages) - 1, -1, -1):
            if isinstance(messages[i], AssistantResponseLLMMessage):
                assistant_count += 1
                if assistant_count >= self.config.min_preserved_assistant_messages:
                    safe_assistant_index = i
                    break

        # If no safe assistant message found, return all messages as full
        if safe_assistant_index == -1:
            return [msg.to_full_message() for msg in messages]

        # Find user message that resulted in assistant response at safe_assistant_index
        # This user message is the one that came before the safe assistant message
        safe_user_index = -1
        for i in range(safe_assistant_index - 1, -1, -1):
            if isinstance(messages[i], UserInputLLMMessage):
                safe_user_index = i
                break

        if safe_user_index == -1:
            # If no safe user message found, compact all messages before the safe assistant message
            safe_index = safe_assistant_index
        else:
            # If safe user message found, make sure it is not compacted
            safe_index = safe_user_index

        # All messages before safe_index are compacted
        result = []
        for i, msg in enumerate(messages):
            if i < safe_index:
                compact_message = msg.to_compact_message()
                if compact_message:
                    result.append(compact_message)
            else:
                result.append(msg.to_full_message())

        return result


# Convenience function for easy integration
def compact_llm_messages(
    messages: List[LLMMessage], config: Optional[CompactionConfig] = None
) -> List[Dict[str, Any]]:
    """Compact a list of LLM messages using the default compactor.

    Args:
        messages: List of LLM messages to compact
        config: Optional compaction configuration

    Returns:
        List of compacted messages in LLM API format
    """
    compactor = LLMContextCompactor(config)
    return compactor.compact_messages(messages)



def simple_shorten(text, width, placeholder="..."):
    if len(text) <= width:
        return text
    return text[: width - len(placeholder)] + placeholder


def to_camel_case(name):
    """
    Convert a string to CamelCase if it's not already.
    Handles snake_case, kebab-case, space-separated, and mixed cases.
    """
    # Check if already in CamelCase
    if is_camel_case(name):
        return name

    # Replace common separators with spaces
    name = name.replace("_", " ").replace("-", " ")

    # Split into words and capitalize each
    words = name.split()

    # Handle empty string
    if not words:
        return name

    # Convert to CamelCase
    camel_case = "".join(word.capitalize() for word in words)

    return camel_case


def is_camel_case(name):
    """
    Check if a string is already in CamelCase format.
    """
    # CamelCase criteria:
    # - No spaces, underscores, or hyphens
    # - Starts with uppercase letter
    # - Contains at least one letter

    if not name or not name[0].isupper():
        return False

    if " " in name or "_" in name or "-" in name:
        return False

    # Check if it has proper capitalization pattern
    # (optional check for strict CamelCase)
    return name.isalnum()



"""
Utility functions for error handling and visibility across the playbooks framework.
"""

import logging
from typing import List, Dict, Any
from rich.console import Console
from rich.panel import Panel
from rich.text import Text

logger = logging.getLogger(__name__)
console = Console()


def print_agent_errors_summary(
    errors: List[Dict[str, Any]], title: str = "Agent Errors"
) -> None:
    """Print a formatted summary of agent errors for CLI visibility.

    Args:
        errors: List of error dictionaries from get_agent_errors()
        title: Title for the error summary panel
    """
    if not errors:
        return

    error_text = Text()
    error_text.append(f"Found {len(errors)} agent error(s):\n\n", style="bold red")

    for i, error_info in enumerate(errors, 1):
        error_text.append(f"{i}. ", style="bold")
        error_text.append(f"Agent {error_info['agent_name']}", style="cyan")
        error_text.append(f" ({error_info['agent_id']})\n", style="dim cyan")
        error_text.append("   Error: ", style="bold")
        error_text.append(f"{error_info['error_type']}: ", style="yellow")
        error_text.append(f"{error_info['error']}\n\n", style="red")

    error_text.append(
        "💡 Check logs with --verbose for full stack traces", style="dim yellow"
    )

    panel = Panel(error_text, title=f"⚠️  {title}", border_style="red", expand=False)
    console.print(panel)


def log_agent_errors(errors: List[Dict[str, Any]], context: str = "execution") -> None:
    """Log agent errors to structured logging for debugging.

    Args:
        errors: List of error dictionaries from get_agent_errors()
        context: Context where the errors occurred (e.g., "execution", "test", "initialization")
    """
    if not errors:
        return

    logger.error(f"Agent errors detected during {context}: {len(errors)} error(s)")

    for error_info in errors:
        logger.error(
            f"Agent error - Agent: {error_info['agent_name']} ({error_info['agent_id']}), "
            f"Error: {error_info['error_type']}: {error_info['error']}",
            extra={
                "agent_id": error_info["agent_id"],
                "agent_name": error_info["agent_name"],
                "error_type": error_info["error_type"],
                "error_message": error_info["error"],
                "context": context,
            },
            exc_info=error_info.get("error_obj"),
        )


def raise_on_agent_errors(
    errors: List[Dict[str, Any]], context: str = "execution"
) -> None:
    """Raise an exception if agent errors are present, for use in tests and strict contexts.

    Args:
        errors: List of error dictionaries from get_agent_errors()
        context: Context where the errors occurred

    Raises:
        RuntimeError: If any agent errors are present
    """
    if not errors:
        return

    error_summary = f"Agent errors detected during {context}:\n"
    for error_info in errors:
        error_summary += f"  - Agent {error_info['agent_name']}: {error_info['error_type']} - {error_info['error']}\n"

    raise RuntimeError(error_summary.strip())


def check_playbooks_health(
    playbooks,
    print_errors: bool = True,
    log_errors: bool = True,
    raise_on_errors: bool = False,
    context: str = "execution",
) -> Dict[str, Any]:
    """Comprehensive health check for a Playbooks instance.

    Args:
        playbooks: Playbooks instance to check
        print_errors: Whether to print errors to console
        log_errors: Whether to log errors to structured logging
        raise_on_errors: Whether to raise an exception if errors are found
        context: Context for error reporting

    Returns:
        Dictionary with health status and error information

    Raises:
        RuntimeError: If raise_on_errors=True and errors are found
    """
    if not playbooks or not playbooks.program:
        return {
            "status": "not_initialized",
            "has_errors": False,
            "error_count": 0,
            "errors": [],
        }

    health = playbooks.check_execution_health()

    if health["has_errors"]:
        errors = health["errors"]

        if log_errors:
            log_agent_errors(errors, context)

        if print_errors:
            print_agent_errors_summary(
                errors, f"Playbooks Health Check - {context.title()}"
            )

        if raise_on_errors:
            raise_on_agent_errors(errors, context)

    return health


class PlaybooksErrorChecker:
    """Context manager for automatic error checking in playbooks execution.

    Usage:
        with PlaybooksErrorChecker(playbooks, raise_on_errors=True):
            await playbooks.program.run_till_exit()
    """

    def __init__(
        self,
        playbooks,
        print_errors: bool = True,
        log_errors: bool = True,
        raise_on_errors: bool = False,
        context: str = "execution",
    ):
        self.playbooks = playbooks
        self.print_errors = print_errors
        self.log_errors = log_errors
        self.raise_on_errors = raise_on_errors
        self.context = context

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        check_playbooks_health(
            self.playbooks,
            print_errors=self.print_errors,
            log_errors=self.log_errors,
            raise_on_errors=self.raise_on_errors,
            context=self.context,
        )
        return False  # Don't suppress exceptions



"""
Utility functions for file handling and type detection.
"""

from pathlib import Path
from typing import List


def is_compiled_playbook_file(file_path: str) -> bool:
    """
    Check if a file is a compiled playbook (.pbasm) file.

    Args:
        file_path: Path to the file to check

    Returns:
        bool: True if the file is a .pbasm file, False otherwise
    """
    return Path(file_path).suffix.lower() == ".pbasm"


def is_source_playbook_file(file_path: str) -> bool:
    """
    Check if a file is a source playbook (.pb or .playbooks) file.

    Args:
        file_path: Path to the file to check

    Returns:
        bool: True if the file is a source playbook file, False otherwise
    """
    suffix = Path(file_path).suffix.lower()
    return suffix in [".pb", ".playbooks"]


def is_playbook_file(file_path: str) -> bool:
    """
    Check if a file is any type of playbook file (.pb, .pbasm, .playbooks).

    Args:
        file_path: Path to the file to check

    Returns:
        bool: True if the file is a playbook file, False otherwise
    """
    return is_source_playbook_file(file_path) or is_compiled_playbook_file(file_path)


def has_compiled_playbook_files(file_paths: List[str]) -> bool:
    """
    Check if any of the provided file paths are compiled playbook files.

    Args:
        file_paths: List of file paths to check

    Returns:
        bool: True if any file is a .pbasm file, False otherwise
    """
    return any(is_compiled_playbook_file(path) for path in file_paths)


def get_file_type_description(file_path: str) -> str:
    """
    Get a human-readable description of the file type.

    Args:
        file_path: Path to the file

    Returns:
        str: Description of the file type
    """
    if is_compiled_playbook_file(file_path):
        return "compiled playbook"
    elif is_source_playbook_file(file_path):
        return "source playbook"
    else:
        return "unknown file type"


def read_file(file_path: str) -> str:
    """
    Read a file and return its content.
    """
    with open(file_path, "r") as file:
        return file.read()



"""
MessagingMixin for event-driven message processing.
"""

import asyncio
import time
from typing import List

from ..constants import EOM, EXECUTION_FINISHED
from ..debug_logger import debug
from ..exceptions import ExecutionFinished
from ..llm_messages import AgentCommunicationLLMMessage
from ..message import Message


class MessagingMixin:
    """Mixin for event-driven message processing functionality."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._message_buffer: List[Message] = []
        self._message_event = asyncio.Event()

    async def _add_message_to_buffer(self, message) -> None:
        """Add a message to buffer and notify waiting processes.

        This is the single entry point for all incoming messages.
        """
        if hasattr(self, "meeting_manager") and self.meeting_manager:
            debug(f"{str(self)}: Adding message to meeting manager: {message}")
            message_handled = await self.meeting_manager._add_message_to_buffer(message)
            if message_handled:
                return

        # Regular messages go to buffer
        debug(f"{str(self)}: Adding message to buffer: {message}")
        self._message_buffer.append(message)
        # Wake up any agents waiting for messages
        debug(f"{str(self)}: Waking up any threads waiting for messages")
        self._message_event.set()

    async def WaitForMessage(self, wait_for_message_from: str) -> List[Message]:
        """Unified message waiting with smart buffering.

        Args:
            wait_for_message_from: Message source - "*", "human", "agent 1234", or "meeting 123"

        Returns:
            Collected messages as string
        """
        while True:
            debug(f"{str(self)}: Waiting for message from {wait_for_message_from}")
            if self.program.execution_finished:
                raise ExecutionFinished(EXECUTION_FINISHED)

            first_message_time = None
            buffer_timeout = 5.0  # 5s maximum buffer time

            release_buffer = False
            num_messages_to_process = 0
            # Check buffer for messages
            for message in self._message_buffer:
                # Track timing from first message
                if first_message_time is None:
                    first_message_time = message.created_at.timestamp()

                # Check if we should release the buffer
                debug(
                    f"{str(self)}: Checking if we should release the buffer for message {message}"
                )
                if release_buffer or self._should_release_buffer(
                    wait_for_message_from, message, first_message_time, buffer_timeout
                ):
                    release_buffer = True
                    num_messages_to_process += 1

                if message.content == EOM:
                    release_buffer = True
                    break
            if release_buffer:
                debug(f"{str(self)}: Releasing buffer")
                return await self._process_collected_messages(num_messages_to_process)

            # Wait for new messages or timeout
            try:
                debug(f"{str(self)}: Waiting for new messages")
                await asyncio.wait_for(self._message_event.wait(), timeout=5)
                debug(f"{str(self)}: New messages received")
                self._message_event.clear()
            except asyncio.TimeoutError:
                # Loop back to process received messages
                debug(f"{str(self)}: Timeout waiting for new messages")
                pass

    def _should_release_buffer(
        self,
        source: str,
        message: Message,
        first_message_time: float,
        buffer_timeout: float,
    ) -> bool:
        """Determine if we should release the buffer now.

        Args:
            source: The source we're waiting for ("human", "agent 1234", "meeting 123")
            message: The message that just arrived
            first_message_time: When we started buffering
            buffer_timeout: Maximum buffer time (5s)

        Returns:
            True if buffer should be released now
        """
        time_elapsed = time.time() - first_message_time if first_message_time else 0

        if source.startswith("meeting "):
            # if the message mentions me, release immediately
            if self.id.lower() in message.content.lower():
                return True

            if self.name.lower() in message.content.lower():
                return True

            # Meeting: always wait full 5s to accumulate chatter
            return time_elapsed >= buffer_timeout
        else:
            # Human/Agent: release immediately on target source OR 5s timeout
            target_source_message = (
                message.sender_id == source
                or message.sender_id == "human"
                or source == "*"
            )
            sent_directly_to_me = message.recipient_id == self.id
            if (
                target_source_message
                or sent_directly_to_me
                or message.content == EOM
                or time_elapsed >= buffer_timeout
            ):
                return True

    async def _process_collected_messages(
        self, num_messages_to_process: int = None
    ) -> List[Message]:
        """Process and format collected messages.

        Args:
            messages: List of message objects

        Returns:
            Formatted message string
        """
        if not num_messages_to_process:
            num_messages_to_process = len(self._message_buffer)

        debug(f"{str(self)}: Processing {num_messages_to_process} messages")
        if not num_messages_to_process:
            return ""

        # Filter out EOM messages before processing
        messages = [
            msg
            for msg in self._message_buffer[:num_messages_to_process]
            if msg.content != EOM
        ]

        # Remove processed messages from buffer
        self._message_buffer = self._message_buffer[num_messages_to_process:]

        if self.state.call_stack.is_empty():
            await self.execute_playbook("ProcessMessages", messages=messages)
        else:
            messages_str = []
            for message in messages:
                messages_str.append(
                    f"Received message from {message.sender_klass}(agent {message.sender_id}): {message.content}"
                )
            debug(f"{str(self)}: Messages to process: {messages_str}")
            # Use the first sender agent for the semantic message type
            sender_agent = messages[0].sender_klass if messages else None
            agent_comm_msg = AgentCommunicationLLMMessage(
                "\n".join(messages_str),
                sender_agent=sender_agent,
                target_agent=self.klass,
            )
            self.state.call_stack.add_llm_message(agent_comm_msg)

        return messages



"""
Playbooks configuration loader (TOML, env, CLI overrides)

Locations (same filename for consistency):
- Project defaults (checked in):        ./playbooks.toml
- User overrides (per-dev, cross-OS):   $XDG_CONFIG_HOME/playbooks/playbooks.toml
  (resolved via platformdirs; Linux: $HOME/.config/playbooks, macOS: ~/Library/Application Support/playbooks/, Windows: %APPDATA%\\playbooks\\)

Profiles:
- Optional sibling files named playbooks.<profile>.toml next to the base file.

Precedence (last wins):
project base < project profile < user base < user profile < explicit --config/PLAYBOOKS_CONFIG < env < CLI overrides

Notes:
- Secrets should come from env (or a secret manager injecting env). Do not store secrets in TOML files.
- Env overrides use PLAYBOOKS_ prefix and “__” for nesting, e.g., PLAYBOOKS_MODEL__TEMPERATURE=0.7
- Python 3.11+ is assumed (tomllib). For 3.10, install tomli and keep the import fallback below.
"""

from __future__ import annotations

import json
import os
import sys
from pathlib import Path
from typing import Any, Iterable, Tuple

import tomllib
from platformdirs import PlatformDirs
from pydantic import BaseModel, ConfigDict, Field, ValidationError

# ---------- Typed schema ----------


class ModelConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")  # catch typos early

    provider: str = "anthropic"
    name: str = "claude-sonnet-4-20250514"
    temperature: float = Field(0.2, ge=0, le=2.0)


class ModelsConfig(BaseModel):
    model_config = ConfigDict(extra="allow")  # allow user-defined models

    execution: ModelConfig | None = None
    compilation: ModelConfig | None = None
    default: ModelConfig | None = None  # fallback model from [model] section

    def model_post_init(self, _):
        if self.default is None:
            self.default = ModelConfig()
        if self.execution is None:
            self.execution = self.default
        if self.compilation is None:
            self.compilation = self.default


class LLMCacheConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")  # catch typos early

    type: str = "disk"  # "disk" or "redis"
    enabled: bool = True
    path: str = ".llm_cache"  # for disk cache


class LangfuseConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")  # catch typos early

    enabled: bool = True


class LitellmConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")  # catch typos early

    verbose: bool = False


class PlaybooksConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")  # catch typos early

    timeout_s: int = 60
    debug: bool = False
    model: ModelsConfig = ModelsConfig()
    llm_cache: LLMCacheConfig = LLMCacheConfig()
    langfuse: LangfuseConfig = LangfuseConfig()
    litellm: LitellmConfig = LitellmConfig()

    def as_dict(self) -> dict[str, Any]:
        return self.model_dump()

    def to_json(self) -> str:
        return self.model_dump_json(indent=2)


# ---------- Paths & discovery ----------

APPNAME = "playbooks"
FILENAME = "playbooks.toml"


def project_cfg_file(cwd: Path | None = None) -> Path:
    root = cwd or Path.cwd()
    return root / FILENAME


def user_cfg_file(user_dir: Path | None = None) -> Path:
    """Return user config file path, allowing an explicit override for tests."""
    if user_dir is not None:
        return Path(user_dir) / FILENAME
    # `appauthor=False` yields ~/.config/playbooks on Linux
    dirs = PlatformDirs(appname=APPNAME, appauthor=False)
    return Path(dirs.user_config_dir) / FILENAME


def profile_variant(base: Path, profile: str | None) -> Path | None:
    if not profile:
        return None
    prof = base.with_name(f"{base.stem}.{profile}{base.suffix}")
    return prof if prof.exists() else None


def _load_toml(path: Path) -> dict[str, Any]:
    with path.open("rb") as f:
        data = tomllib.load(f) or {}
    if not isinstance(data, dict):
        raise ValueError(f"{path} must contain a top-level TOML table (mapping).")
    return data


# ---------- Merge & env handling ----------


def deep_merge(a: dict[str, Any], b: dict[str, Any]) -> dict[str, Any]:
    """Deep-merge two mappings (dicts). Lists and scalars are replaced by b."""
    out = dict(a)
    for k, v in b.items():
        if isinstance(v, dict) and isinstance(out.get(k), dict):
            out[k] = deep_merge(out[k], v)  # type: ignore[assignment]
        else:
            out[k] = v
    return out


def _parse_env_value(raw: str) -> Any:
    s = raw.strip()
    # bool/null
    low = s.lower()
    if low in ("true", "false"):
        return low == "true"
    if low in ("null", "none"):
        return None
    # numbers / objects / arrays via JSON if it looks JSON-ish or numeric
    if (s and s[0] in '{["-0123456789') or s in ("0",):
        try:
            return json.loads(s)
        except Exception:
            pass
    return s


def apply_env_overrides(
    base: dict[str, Any], prefix: str = "PLAYBOOKS_", nested_delim: str = "__"
) -> dict[str, Any]:
    """Overlay environment variables onto base dict using nested paths."""
    result = dict(base)
    # Skip profile and environment control variables - they're for config selection, not config content
    skip_keys = {"profile", "environment"}
    for k, v in os.environ.items():
        if not k.startswith(prefix):
            continue
        path = k[len(prefix) :]
        parts = path.split(nested_delim)
        # normalize to lower-case keys to match typical TOML style
        parts = [p.lower() for p in parts if p]
        # Skip profile selection variables
        if parts and parts[0] in skip_keys:
            continue
        _set_by_path(result, parts, _parse_env_value(v))
    return result


def _set_by_path(obj: dict[str, Any], parts: Iterable[str], value: Any) -> None:
    it = iter(parts)
    cur = obj
    try:
        last = next(it)
    except StopIteration:
        return
    for p in it:
        cur = cur.setdefault(last, {})
        if not isinstance(cur, dict):
            raise TypeError(f"Cannot set nested key under non-dict at {last}")
        last = p
    cur.setdefault(last, None)
    cur[last] = value


# ---------- Public API ----------


def resolve_config_files(
    profile: str | None = None,
    explicit_path: str | os.PathLike[str] | None = None,
    cwd: Path | None = None,
    user_config_dir: Path | None = None,
) -> Tuple[Path, ...]:
    """Return the list of config files to load in order (lowest → highest precedence among files)."""
    files: list[Path] = []
    proj = project_cfg_file(cwd)
    if proj.exists():
        files.append(proj)
        if prof := profile_variant(proj, profile):
            files.append(prof)

    user = user_cfg_file(user_config_dir)
    if user.exists():
        files.append(user)
        if prof := profile_variant(user, profile):
            files.append(prof)

    if explicit_path:
        ep = Path(explicit_path).expanduser()
        if ep.exists():
            files.append(ep)

    return tuple(files)


def load_config(
    *,
    profile: str | None = None,
    explicit_path: str | os.PathLike[str] | None = None,
    overrides: dict[str, Any] | None = None,
    cwd: Path | None = None,
    user_config_dir: Path | None = None,
) -> Tuple[PlaybooksConfig, Tuple[Path, ...]]:
    """
    Load config with precedence:
    files (project<profile<user<profile<explicit>) < env(PLAYBOOKS_*) < CLI overrides (passed in `overrides`)
    Returns (config, files_used)
    """
    # Auto-detect profile from environment if not explicitly provided
    if profile is None:
        profile = os.environ.get("PLAYBOOKS_PROFILE") or os.environ.get("ENVIRONMENT")

    files = resolve_config_files(
        profile=profile,
        explicit_path=explicit_path,
        cwd=cwd,
        user_config_dir=user_config_dir,
    )
    merged: dict[str, Any] = {}
    for f in files:
        merged = deep_merge(merged, _load_toml(f))

    merged = apply_env_overrides(merged, prefix="PLAYBOOKS_", nested_delim="__")

    if overrides:
        merged = deep_merge(merged, overrides)

    # Handle [model] section as default fallback before validation
    if "model" in merged and isinstance(merged["model"], dict):
        model_section = merged["model"]
        if "provider" in model_section:
            # This is a direct [model] section with model config
            # Extract only ModelConfig fields for the default
            default_model = {
                k: v
                for k, v in model_section.items()
                if k in ["provider", "name", "temperature"] and not isinstance(v, dict)
            }
            merged["model"]["default"] = default_model

    try:
        config = PlaybooksConfig.model_validate(merged)
        return config, files
    except ValidationError as e:
        # pretty print and exit if used as a CLI helper
        print(e, file=sys.stderr)
        raise


# Global config instance
config, _ = load_config()


__all__ = [
    "PlaybooksConfig",
    "ModelConfig",
    "ModelsConfig",
    "LLMCacheConfig",
    "LangfuseConfig",
    "config",
    "load_config",
    "resolve_config_files",
    "deep_merge",
    "apply_env_overrides",
]



from typing import Any, Dict, List, Optional

from playbooks.llm_messages import LLMMessage

from .event_bus import EventBus
from .events import CallStackPopEvent, CallStackPushEvent, InstructionPointerEvent
from .playbook_step import PlaybookStep


class InstructionPointer:
    """Represents a position in a playbook.

    Attributes:
        playbook: The name of the playbook.
        line_number: The line number within the playbook.
        source_line_number: The source line number in the markdown.
        source_file_path: The file path of the source markdown.
    """

    def __init__(
        self,
        playbook: str,
        line_number: str,
        source_line_number: int,
        step: PlaybookStep = None,
        source_file_path: str = None,
    ):
        self.playbook = playbook
        self.line_number = line_number
        self.source_line_number = source_line_number
        self.source_file_path = source_file_path
        self.step = step

    def copy(self) -> "InstructionPointer":
        return InstructionPointer(
            self.playbook,
            self.line_number,
            self.source_line_number,
            self.step,
            self.source_file_path,
        )

    def increment_instruction_pointer(self) -> None:
        # TODO: this is a hack to advance the instruction pointer
        self.line_number = str(int(self.line_number) + 1)
        self.source_line_number = self.source_line_number + 1

    def to_compact_str(self) -> str:
        compact_str = (
            self.playbook
            if self.line_number is None
            else f"{self.playbook}:{self.line_number}"
        )
        return compact_str

    def __str__(self) -> str:
        compact_str = self.to_compact_str()
        if self.source_line_number is not None:
            return f"{compact_str} ({self.source_file_path}:{self.source_line_number})"
        return compact_str

    def __repr__(self) -> str:
        return str(self)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "playbook": self.playbook,
            "line_number": self.line_number,
            "source_line_number": self.source_line_number,
        }


class CallStackFrame:
    """Represents a frame in the call stack.

    Attributes:
        instruction_pointer: Points to the current instruction.
        llm_chat_session_id: ID of the associated LLM chat session, if any.
    """

    def __init__(
        self,
        instruction_pointer: InstructionPointer,
        llm_messages: Optional[List[LLMMessage]] = None,
        langfuse_span: Optional[Any] = None,
        is_meeting: bool = False,
        meeting_id: Optional[str] = None,
    ):
        self.instruction_pointer = instruction_pointer
        self.llm_messages = llm_messages or []
        self.langfuse_span = langfuse_span
        self.is_meeting = is_meeting
        self.meeting_id = meeting_id
        self.depth = -1

    @property
    def source_line_number(self) -> int:
        return self.instruction_pointer.source_line_number

    @property
    def line_number(self) -> int:
        return self.instruction_pointer.line_number

    @property
    def playbook(self) -> str:
        return self.instruction_pointer.playbook

    @property
    def step(self) -> PlaybookStep:
        return self.instruction_pointer.step

    def to_dict(self) -> Dict[str, Any]:
        """Convert the frame to a dictionary representation.

        Returns:
            A dictionary representation of the frame.
        """
        result = {
            "instruction_pointer": str(self.instruction_pointer),
            "langfuse_span": str(self.langfuse_span) if self.langfuse_span else None,
        }
        if self.is_meeting:
            result["is_meeting"] = self.is_meeting
            result["meeting_id"] = self.meeting_id
        return result

    def add_llm_message(self, message: LLMMessage) -> None:
        """Add an LLMMessage object to the call stack frame."""
        self.llm_messages.append(message)

    def __repr__(self) -> str:
        base_repr = self.instruction_pointer.to_compact_str()
        if self.is_meeting and self.meeting_id:
            return f"{base_repr}[meeting {self.meeting_id}]"
        return base_repr

    def get_llm_messages(self) -> List[Dict[str, str]]:
        """Get the messages for the call stack frame as dictionaries for LLM API."""
        return [msg.to_full_message() for msg in self.llm_messages]


class CallStack:
    """A stack of call frames."""

    def __init__(self, event_bus: EventBus, agent_id: str = "unknown"):
        self.frames: List[CallStackFrame] = []
        self.event_bus = event_bus
        self.agent_id = agent_id

    def is_empty(self) -> bool:
        """Check if the call stack is empty.

        Returns:
            True if the call stack has no frames, False otherwise.
        """
        return not self.frames

    def push(self, frame: CallStackFrame) -> None:
        """Push a frame onto the call stack.

        Args:
            frame: The frame to push.
        """
        self.frames.append(frame)
        frame.depth = len(self.frames)
        event = CallStackPushEvent(
            session_id=self.agent_id, frame=str(frame), stack=self.to_dict()
        )
        self.event_bus.publish(event)

    def pop(self) -> Optional[CallStackFrame]:
        """Remove and return the top frame from the call stack.

        Returns:
            The top frame, or None if the stack is empty.
        """
        frame = self.frames.pop() if self.frames else None
        if frame:
            event = CallStackPopEvent(
                session_id=self.agent_id, frame=str(frame), stack=self.to_dict()
            )
            self.event_bus.publish(event)
        return frame

    def peek(self) -> Optional[CallStackFrame]:
        """Return the top frame without removing it.

        Returns:
            The top frame, or None if the stack is empty.
        """
        return self.frames[-1] if self.frames else None

    def advance_instruction_pointer(
        self, instruction_pointer: InstructionPointer
    ) -> None:
        """Advance the instruction pointer to the next instruction.

        Args:
            instruction_pointer: The new instruction pointer.
        """
        self.frames[-1].instruction_pointer = instruction_pointer
        event = InstructionPointerEvent(
            session_id=self.agent_id,
            pointer=str(instruction_pointer),
            stack=self.to_dict(),
        )
        self.event_bus.publish(event)

    def __repr__(self) -> str:
        frames = ", ".join(str(frame.instruction_pointer) for frame in self.frames)
        return f"CallStack[{frames}]"

    def __str__(self) -> str:
        return self.__repr__()

    def to_dict(self) -> List[str]:
        """Convert the call stack to a dictionary representation.

        Returns:
            A list of string representations of instruction pointers.
        """
        return [frame.instruction_pointer.to_dict() for frame in self.frames]

    def get_llm_messages(self) -> List[Dict[str, str]]:
        """Get the messages for the call stack for the LLM."""
        messages = []
        for frame in self.frames:
            messages.extend(frame.get_llm_messages())
        return messages

    def add_llm_message(self, message) -> None:
        """Safely add an LLM message to the top frame if the stack is not empty."""
        current_frame = self.peek()
        if current_frame is not None:
            current_frame.add_llm_message(message)



import hashlib
import logging
import os
import tempfile
import time
from functools import wraps
from typing import Any, Callable, Iterator, List, Optional, TypeVar, Union

import litellm
from litellm import completion, get_supported_openai_params

from playbooks.enums import LLMMessageRole
from playbooks.llm_messages import SystemPromptLLMMessage, UserInputLLMMessage

from ..config import config
from ..constants import SYSTEM_PROMPT_DELIMITER
from ..debug_logger import debug
from ..exceptions import VendorAPIOverloadedError, VendorAPIRateLimitError
from .langfuse_helper import LangfuseHelper
from .llm_config import LLMConfig
from .playbooks_lm_handler import PlaybooksLMHandler

# https://github.com/BerriAI/litellm/issues/2256#issuecomment-2041374430
loggers = ["LiteLLM Proxy", "LiteLLM Router", "LiteLLM"]

for logger_name in loggers:
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.CRITICAL + 1)

litellm.suppress_debug_info = True
# Handle different litellm versions
litellm.drop_params = True
# Note: LLM_API_BASE is now applied per-model basis, not globally
# litellm._turn_on_debug()

# Initialize the Playbooks-LM handler
playbooks_handler = PlaybooksLMHandler()

# Store the original completion function
_original_completion = completion


def completion_with_preprocessing(*args, **kwargs):
    """
    Wrapper for litellm.completion that applies preprocessing for playbooks-lm models.
    """
    model = kwargs.get("model", "")

    # Note: Playbooks-LM preprocessing is now handled in get_completion() before Langfuse logging

    # Debug: log the call to help diagnose auth issues when verbose mode is enabled
    if os.getenv("LLM_SET_VERBOSE", "False").lower() == "true":
        api_key_preview = kwargs.get("api_key", "MISSING")
        if api_key_preview and api_key_preview != "MISSING":
            api_key_preview = (
                api_key_preview[:8] + "..." if len(api_key_preview) > 8 else "short"
            )
        debug(
            "LLM Call",
            model=model,
            api_base=kwargs.get("api_base", "default"),
            api_key_preview=api_key_preview,
        )

    # Call the original completion function
    return _original_completion(*args, **kwargs)


# Replace litellm's completion function with our wrapper
litellm.completion = completion_with_preprocessing
completion = completion_with_preprocessing

# Initialize cache if enabled
cache = None

# Load cache configuration from config system with environment fallback
llm_cache_enabled = config.llm_cache.enabled
llm_cache_type = config.llm_cache.type.lower()
llm_cache_path = config.llm_cache.path

if llm_cache_enabled:
    if llm_cache_type == "disk":
        from diskcache import Cache

        cache_dir = (
            llm_cache_path or tempfile.TemporaryDirectory(prefix="llm_cache_").name
        )
        cache = Cache(directory=cache_dir)

    elif llm_cache_type == "redis":
        from redis import Redis

        redis_url = os.getenv("REDIS_URL", "redis://127.0.0.1:6379/0")
        cache = Redis.from_url(redis_url)
        debug("Using LLM cache", redis_url=redis_url)

    else:
        raise ValueError(f"Invalid LLM cache type: {llm_cache_type}")


def custom_get_cache_key(**kwargs) -> str:
    """Generate a deterministic cache key based on request parameters.

    Args:
        **kwargs: The completion request parameters

    Returns:
        A unique hash string to use as cache key
    """
    import json

    # Create a deterministic representation of the cache key components
    cache_components = {
        "model": kwargs.get("model", ""),
        "messages": kwargs.get("messages", []),
        "temperature": kwargs.get("temperature", 0.2),
        "logit_bias": kwargs.get("logit_bias", {}),
    }

    # Use json.dumps with sort_keys=True for deterministic serialization
    key_str = json.dumps(cache_components, sort_keys=True, separators=(",", ":"))
    return hashlib.sha256(key_str.encode("utf-8")).hexdigest()[:32]


T = TypeVar("T")


def retry_on_overload(
    max_retries: int = 3, base_delay: float = 1.0
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """Decorator that retries a function on API overload or rate limit errors with exponential backoff.

    Args:
        max_retries: Maximum number of retry attempts
        base_delay: Initial delay between retries in seconds

    Returns:
        A decorator function that adds retry logic
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T:
            for attempt in range(max_retries):
                try:
                    return func(*args, **kwargs)
                except (
                    VendorAPIOverloadedError,
                    VendorAPIRateLimitError,
                    litellm.RateLimitError,
                    litellm.InternalServerError,
                    litellm.ServiceUnavailableError,
                    litellm.APIConnectionError,
                    litellm.Timeout,
                ):
                    if attempt == max_retries - 1:
                        # Last attempt, re-raise the exception
                        raise

                    delay = base_delay * (2**attempt)
                    time.sleep(delay)
                    continue
            return func(*args, **kwargs)  # This line should never be reached

        return wrapper

    return decorator


@retry_on_overload()
def _make_completion_request(completion_kwargs: dict) -> str:
    """Make a non-streaming completion request to the LLM with automatic retries on overload."""
    response = completion(**completion_kwargs)
    return response["choices"][0]["message"]["content"]


def _make_completion_request_stream(completion_kwargs: dict) -> Iterator[str]:
    """Make a streaming completion request to the LLM with automatic retries on overload.

    Since exceptions occur on the first token (during initial call), we can retry
    the entire stream creation and maintain true streaming.
    """
    max_retries = 5
    base_delay = 1.0

    for attempt in range(max_retries):
        try:
            response = completion(**completion_kwargs)

            # Try to get the first chunk to trigger any immediate exceptions
            first_chunk = None
            response_iter = iter(response)
            try:
                first_chunk = next(response_iter)
            except StopIteration:
                # Empty response
                return

            # Yield the first chunk
            content = first_chunk.choices[0].delta.content
            if content is not None:
                yield content

            # Now stream the rest normally
            for chunk in response_iter:
                content = chunk.choices[0].delta.content
                if content is not None:
                    yield content

            return  # Success, exit retry loop

        except (
            VendorAPIOverloadedError,
            VendorAPIRateLimitError,
            litellm.RateLimitError,
            litellm.InternalServerError,
            litellm.ServiceUnavailableError,
            litellm.APIConnectionError,
            litellm.Timeout,
        ):
            if attempt == max_retries - 1:
                # Last attempt, re-raise the exception
                raise

            delay = base_delay * (2**attempt)
            time.sleep(delay)
            continue


def get_completion(
    llm_config: LLMConfig,
    messages: List[dict],
    stream: bool = False,
    use_cache: bool = True,
    json_mode: bool = False,
    session_id: Optional[str] = None,
    langfuse_span: Optional[Any] = None,
    **kwargs,
) -> Iterator[str]:
    """Get completion from LLM with optional streaming and caching support.

    Args:
        llm_config: LLM configuration containing model and API key
        messages: List of message dictionaries to send to the LLM
        stream: If True, returns an iterator of response chunks
        use_cache: If True and caching is enabled, will try to use cached responses
        json_mode: If True, instructs the model to return a JSON response
        session_id: Optional session ID to associate with the generation
        langfuse_span: Optional parent span for Langfuse tracing
        **kwargs: Additional arguments passed to litellm.completion

    Returns:
        An iterator of response text (single item for non-streaming)
    """

    messages = remove_empty_messages(messages)
    messages = consolidate_messages(messages)
    messages = ensure_upto_N_cached_messages(messages)

    # Apply playbooks-lm preprocessing if needed (before Langfuse logging)
    if "playbooks-lm" in llm_config.model.lower():
        messages = playbooks_handler.preprocess_messages(messages.copy())

    completion_kwargs = {
        "model": llm_config.model,
        "api_key": llm_config.api_key,
        "messages": messages.copy(),
        "max_completion_tokens": 7500,
        "stream": stream,
        "temperature": llm_config.temperature,
        **kwargs,
    }

    # Add response_format for JSON mode if supported by the model
    if json_mode:
        params = get_supported_openai_params(model=llm_config.model)
        if "response_format" in params:
            completion_kwargs["response_format"] = {"type": "json_object"}

    # Initialize Langfuse tracing if available
    langfuse_span_obj = None
    if langfuse_span is None:
        langfuse_helper = LangfuseHelper.instance()
        if langfuse_helper is not None:
            langfuse_span_obj = langfuse_helper.trace(
                name="llm_call",
                metadata={"model": llm_config.model, "session_id": session_id},
            )
    else:
        langfuse_span_obj = langfuse_span

    langfuse_generation = None
    if langfuse_span_obj is not None:
        langfuse_generation = langfuse_span_obj.generation(
            model=llm_config.model,
            model_parameters={
                "maxTokens": completion_kwargs["max_completion_tokens"],
                "temperature": completion_kwargs["temperature"],
            },
            input=messages,
            session_id=session_id,
            metadata={"stream": stream},
        )

    # Try to get response from cache if enabled
    if llm_cache_enabled and use_cache and cache is not None:
        cache_key = custom_get_cache_key(**completion_kwargs)
        cache_value = cache.get(cache_key)

        if cache_value is not None:
            if langfuse_generation:
                langfuse_generation.update(metadata={"cache_hit": True})
                langfuse_generation.end(output=str(cache_value))
                langfuse_generation.update(cost_details={"input": 0, "output": 0})
                LangfuseHelper.flush()

            if stream:
                for chunk in cache_value:
                    yield chunk
            else:
                yield cache_value

            return

    # Get response from LLM
    full_response: Union[str, List[str]] = [] if stream else ""
    error_occurred = False
    try:
        if langfuse_generation:
            langfuse_generation.update(metadata={"cache_hit": False})

        if stream:
            for chunk in _make_completion_request_stream(completion_kwargs):
                full_response.append(chunk)  # type: ignore
                yield chunk
            full_response = "".join(full_response)  # type: ignore
        else:
            full_response = _make_completion_request(completion_kwargs)
            yield full_response
    except Exception as e:
        error_occurred = True
        if langfuse_generation:
            langfuse_generation.end(error=str(e))
            LangfuseHelper.flush()
        raise e  # Re-raise the exception to be caught by the decorator if applicable
    finally:
        # Update cache and Langfuse
        if (
            not error_occurred
            and llm_cache_enabled
            and use_cache
            and cache is not None
            and full_response is not None
            and len(full_response) > 0
        ):
            full_response = str(full_response)
            cache.set(cache_key, full_response)

        if langfuse_generation and not error_occurred:
            # Only update if no exception occurred
            langfuse_generation.end(output=full_response)
            LangfuseHelper.flush()


def remove_empty_messages(messages: List[dict]) -> List[dict]:
    """Remove empty messages from the list."""
    return [message for message in messages if message["content"].strip()]


def get_messages_for_prompt(prompt: str) -> List[dict]:
    """Convert a raw prompt into a properly formatted message list.

    If the prompt contains a system prompt delimiter, it will be split into
    separate system and user messages. Otherwise, treated as a system message.

    Args:
        prompt: The raw prompt text, potentially containing a system/user split

    Returns:
        A list of message dictionaries formatted for LLM API calls
    """
    if SYSTEM_PROMPT_DELIMITER in prompt:
        system, user = prompt.split(SYSTEM_PROMPT_DELIMITER)

        return [
            SystemPromptLLMMessage(system.strip()).to_full_message(),
            UserInputLLMMessage(user.strip()).to_full_message(),
        ]
    return [UserInputLLMMessage(prompt.strip()).to_full_message()]


def consolidate_messages(messages: List[dict]) -> List[dict]:
    """
    Consolidate consecutive messages where possible
    Break into separate messages for different roles and to include upto 1 cached message
    """

    # First, group messages that can be combined into a single message
    message_groups = []
    current_group = []
    current_role = messages[0]["role"]

    for message in messages:
        if "cache_control" in message and message["role"] == current_role:
            # Include the cached message in the current group
            current_group.append(message)
            message_groups.append(current_group)

            # Start a new group
            current_group = []
        elif message["role"] == current_role:
            current_group.append(message)
        else:
            # New role, so start a new group with this message in it
            message_groups.append(current_group)
            current_group = [message]
            current_role = message["role"]

    if current_group:
        message_groups.append(current_group)

    # Now, consolidate each group into a single message
    messages = []
    for group in message_groups:
        if not group:
            continue
        contents = []
        cache_control = False

        # Collect all contents and track if there is a cached message
        for message in group:
            contents.append(message["content"])
            if "cache_control" in message:
                cache_control = True

        # Join all contents into a single string
        contents = "\n\n".join(contents)

        # Add the consolidated message to the list
        from playbooks.llm_messages import LLMMessage

        if cache_control:
            llm_msg = LLMMessage(
                contents, LLMMessageRole(group[0]["role"]), cached=True
            )
        else:
            llm_msg = LLMMessage(
                contents, LLMMessageRole(group[0]["role"]), cached=False
            )
        messages.append(llm_msg.to_full_message())

    return messages


def ensure_upto_N_cached_messages(messages: List[dict]) -> List[dict]:
    """
    Ensure that there are at most N cached messages in the list.
    If there are more than N, keep the last N.
    """

    max_cached_messages = 4 - 1  # Keep one for the System message
    count_cached_messages = 0

    # Cached messages are those with a cache_control field set
    # Scan in reverse order to keep the last N cached messages
    for message in reversed(messages):
        # If we've already found N cached messages, remove cache_control from all earlier messages
        if count_cached_messages >= max_cached_messages:
            # Don't remove cache_control from the System message
            if message["role"] == LLMMessageRole.SYSTEM:
                continue

            # Remove cache_control from all other messages
            if "cache_control" in message:
                del message["cache_control"]

            continue

        # If we haven't found N cached messages yet, check if this message is cached
        if "cache_control" in message:
            count_cached_messages += 1

    return messages



from playbooks.utils.markdown_to_ast import markdown_to_ast

from ..compiler import Compiler
from ..utils.llm_config import LLMConfig


class BuiltinPlaybooks:
    """Provides built-in playbooks that are automatically added to every agent."""

    @staticmethod
    def get_ast_nodes():
        """Get AST nodes for built-in playbooks.

        Returns:
            List of AST nodes representing built-in playbooks.
        """
        return (
            BuiltinPlaybooks.get_python_playbooks_ast_nodes()
            + BuiltinPlaybooks.get_llm_playbooks_ast_nodes()
        )

    @staticmethod
    def get_python_playbooks_ast_nodes():
        """Get AST nodes for built-in python playbooks.

        Returns:
            List of AST nodes representing built-in python playbooks.
        """
        code_block = '''
```python
from playbooks.utils.spec_utils import SpecUtils

@playbook(hidden=True)
async def SendMessage(target_agent_id: str, message: str):
    await agent.SendMessage(target_agent_id, message)

@playbook(hidden=True)
async def WaitForMessage(source_agent_id: str) -> str | None:
    return await agent.WaitForMessage(source_agent_id)

@playbook
async def Say(target: str, message: str, already_streamed: bool = False):
    await agent.Say(target, message, already_streamed)

@playbook
async def CreateAgent(agent_klass: str, **kwargs):
    new_agent = await agent.program.create_agent(agent_klass, **kwargs)
    await agent.program.runtime.start_agent(new_agent)
    return new_agent
    
@playbook
async def SaveArtifact(artifact_name: str, artifact_summary: str, artifact_content: str):
    agent.state.artifacts.set(artifact_name, artifact_summary, artifact_content)

@playbook
async def LoadArtifact(artifact_name: str):
    return agent.state.artifacts[artifact_name]

@playbook
async def InviteToMeeting(meeting_id: str, attendees: list):
    """Invite additional agents to an existing meeting."""
    return await agent.meeting_manager.InviteToMeeting(meeting_id, attendees)

@playbook(hidden=True)
async def AcceptMeetingInvitation(meeting_id: str, inviter_id: str, topic: str, meeting_playbook_name: str):
    return await agent.meeting_manager._accept_meeting_invitation(meeting_id, inviter_id, topic, meeting_playbook_name)

@playbook
async def Loadfile(file_path: str, inline: bool = False, silent: bool = False):
    return await agent.load_file(file_path, inline, silent)

@playbook(hidden=True)
async def SetVar(name: str, value):
    """Set a variable in the agent's state and return the value."""
    if not name.startswith("$"):
        name = f"${name}"
    agent.state.variables[name] = value
    return value

```        
'''

        return markdown_to_ast(code_block)["children"]

    @staticmethod
    def get_llm_playbooks_markdown():
        return """
# BuiltinPlaybooks
## ResolveDescriptionPlaceholders($playbook_call: str, $description: str) -> str
Resolves natural language placeholders as Python expressions in provided playbook description in the context of the provided playbook call
hidden: true

### Steps
- Provided $description contains contains some placeholders in {} in Python f-string syntax
- Go through each placeholder $expression
- If $expression is not valid Python syntax and is a natural language instruction
    - Attempt to convert it to valid Python syntax. If ambiguous or not known how to convert, leave it as is.
- Return description with any converted placeholders. No other changes to description allowed.

## MessageProcessingEventLoop
hidden: true

### Steps
- Loop forever
  - Set $_busy = False
  - WaitForMessage("*")
  - If you received a MEETING INVITATION
    - Find a suitable meeting playbook
    - If meeting playbook is found
      - Set $_busy = True
      - AcceptMeetingInvitation(meeting_id, inviter_id, topic, meeting_playbook_name)
      - Continue
  - Here we will decide if any playbook should be triggered. Carefully consider recent messages received and the current state. Now go through available triggers listed above, and think deeply about whether any should be triggered. 
  - If any playbook should be triggered, set $_busy = True and output an appropriate trig? line.
  - If a playbook was executed, look at the message that was received and the result of the playbook execution. If the message sender is expecting a response, enqueue a Say(message sender, result of the playbook execution) call
"""

    @staticmethod
    def get_llm_playbooks_ast_nodes():
        markdown = BuiltinPlaybooks.get_llm_playbooks_markdown()
        llm_config = LLMConfig()
        compiler = Compiler(llm_config)

        # Compile the playbooks content
        _, compiled_content, compiled_file_path = compiler.compile(content=markdown)

        # Parse the compiled content to extract steps
        ast = markdown_to_ast(compiled_content, source_file_path=compiled_file_path)
        h1 = list(filter(lambda node: node.get("type") == "h1", ast.get("children")))[0]
        if not h1.get("type") == "h1":
            raise Exception("Expected a single h1 child")

        # filter h1 children for h2 nodes
        h2s = list(filter(lambda node: node.get("type") == "h2", h1.get("children")))
        return h2s



"""Utilities for handling agent and meeting specifications and IDs."""


class SpecUtils:
    """Centralized utilities for handling agent and meeting specs/IDs."""

    # Constants for prefixes
    AGENT_PREFIX = "agent "
    MEETING_PREFIX = "meeting "

    @classmethod
    def is_agent_spec(cls, value: str) -> bool:
        """Check if value is an agent specification."""
        return value.startswith(cls.AGENT_PREFIX)

    @classmethod
    def is_meeting_spec(cls, value: str) -> bool:
        """Check if value is a meeting specification."""
        return value.startswith(cls.MEETING_PREFIX)

    @classmethod
    def extract_agent_id(cls, agent_spec: str) -> str:
        """Extract agent ID from agent specification.

        Args:
            agent_spec: Agent specification (e.g., "agent 1000") or raw ID (e.g., "1000")

        Returns:
            Agent ID without prefix (e.g., "1000")
        """
        if cls.is_agent_spec(agent_spec):
            agent_id = agent_spec[len(cls.AGENT_PREFIX) :].strip()
        else:
            agent_id = agent_spec

        if agent_id in ["human", "user", "HUMAN", "USER"]:
            agent_id = "human"

        return agent_id

    @classmethod
    def extract_meeting_id(cls, meeting_spec: str) -> str:
        """Extract meeting ID from meeting specification.

        Args:
            meeting_spec: Meeting specification (e.g., "meeting 123") or raw ID (e.g., "123")

        Returns:
            Meeting ID without prefix (e.g., "123")
        """
        if cls.is_meeting_spec(meeting_spec):
            return meeting_spec[len(cls.MEETING_PREFIX) :].strip()
        return meeting_spec  # Already an ID

    @classmethod
    def to_agent_spec(cls, agent_id: str) -> str:
        """Convert agent ID to agent specification.

        Args:
            agent_id: Agent ID (e.g., "1000") or existing spec (e.g., "agent 1000")

        Returns:
            Agent specification with prefix (e.g., "agent 1000")
        """
        if cls.is_agent_spec(agent_id):
            return agent_id  # Already a spec
        return f"{cls.AGENT_PREFIX}{agent_id}"

    @classmethod
    def to_meeting_spec(cls, meeting_id: str) -> str:
        """Convert meeting ID to meeting specification.

        Args:
            meeting_id: Meeting ID (e.g., "123") or existing spec (e.g., "meeting 123")

        Returns:
            Meeting specification with prefix (e.g., "meeting 123")
        """
        if cls.is_meeting_spec(meeting_id):
            return meeting_id  # Already a spec
        return f"{cls.MEETING_PREFIX}{meeting_id}"



"""
Production logging configuration for Playbooks framework.
This configures Python's standard logging for production use,
separate from the debug_logger used for development.
"""

import logging
import logging.handlers
import os
import sys
import threading
from typing import Optional

from .logging_constants import (
    ENV_LOG_LEVEL,
    ENV_LOG_FILE,
    DEFAULT_LOG_LEVEL,
    DEFAULT_LOG_FORMAT,
    DEBUG_LOG_FORMAT,
)

# Global lock for thread-safe configuration
_config_lock: threading.Lock = threading.Lock()
_configured: bool = False


def setup_production_logging(
    level: Optional[str] = None,
    log_file: Optional[str] = None,
    format_string: Optional[str] = None,
) -> None:
    """
    Configure Python's standard logging for production use.

    This is for structured logging of production events, errors, and metrics.
    It's separate from debug_logger which is for developer troubleshooting.

    Args:
        level: Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)
        log_file: Optional file path for logging output
        format_string: Optional custom format string
    """
    # Get configuration from environment or use defaults
    log_level = level or os.getenv(ENV_LOG_LEVEL, DEFAULT_LOG_LEVEL)
    log_file_path = log_file or os.getenv(ENV_LOG_FILE)

    # Default format for structured logging with context
    if format_string is None:
        format_string = DEFAULT_LOG_FORMAT
        if log_level == "DEBUG":
            # More detailed format for debug mode
            format_string = DEBUG_LOG_FORMAT

    # Configure root logger
    root_logger = logging.getLogger()
    root_logger.setLevel(getattr(logging, log_level.upper()))

    # Thread-safe handler management to avoid duplicates
    # Only clear if handlers exist and we're reconfiguring
    if root_logger.handlers:
        # Make a copy to avoid modification during iteration
        for handler in root_logger.handlers[:]:
            root_logger.removeHandler(handler)
            handler.close()

    # Console handler for stderr (production logs shouldn't interfere with stdout)
    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setFormatter(logging.Formatter(format_string))
    root_logger.addHandler(console_handler)

    # File handler if specified
    if log_file_path:
        # Use rotating file handler to prevent unbounded growth
        file_handler = logging.handlers.RotatingFileHandler(
            log_file_path, maxBytes=10 * 1024 * 1024, backupCount=5  # 10MB
        )
        file_handler.setFormatter(logging.Formatter(format_string))
        root_logger.addHandler(file_handler)

    # Set specific log levels for noisy libraries
    logging.getLogger("httpx").setLevel(logging.WARNING)
    logging.getLogger("httpcore").setLevel(logging.WARNING)
    logging.getLogger("asyncio").setLevel(logging.WARNING)
    logging.getLogger("websockets").setLevel(logging.WARNING)

    # Playbooks-specific loggers can be more verbose
    if log_level == "DEBUG":
        logging.getLogger("playbooks").setLevel(logging.DEBUG)
    else:
        # Even in INFO mode, keep framework internals at INFO level
        logging.getLogger("playbooks").setLevel(logging.INFO)
        # But reduce noise from specific subsystems
        logging.getLogger("playbooks.transport").setLevel(logging.WARNING)
        logging.getLogger("playbooks.debug.server").setLevel(logging.INFO)


def configure_logging() -> None:
    """
    Apply both debug and production logging configuration.
    Call this at application startup.
    Thread-safe: can be called multiple times safely.
    """
    global _configured

    with _config_lock:
        if _configured:
            return  # Already configured, skip

        # Setup production logging
        setup_production_logging()

        # Apply user output configuration
        from playbooks.logging_config import apply_config

        apply_config()

        _configured = True



"""Transport layer for remote playbook execution.

This module provides the transport protocol interface and implementations
for communicating with remote agents and services.
"""

from .mcp_transport import MCPTransport
from .protocol import TransportProtocol

__all__ = ["TransportProtocol", "MCPTransport"]



"""Base LLM execution strategy."""

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from ..playbook_call import PlaybookCall
from ..utils.expression_engine import (
    ExpressionContext,
    resolve_description_placeholders,
)

if TYPE_CHECKING:
    from ..agents.local_ai_agent import LocalAIAgent
    from ..playbook import LLMPlaybook


class LLMExecution(ABC):
    """Base class for LLM execution strategies."""

    def __init__(self, agent: "LocalAIAgent", playbook: "LLMPlaybook"):
        """Initialize the execution strategy.

        Args:
            agent: The agent executing the playbook
            playbook: The LLM playbook to execute
        """
        self.agent: "LocalAIAgent" = agent
        self.playbook: "LLMPlaybook" = playbook

    async def resolve_description_placeholders(
        self, description: str, *args, **kwargs
    ) -> str:
        """Resolve description placeholders if present.

        Args:
            description: The description potentially containing placeholders
            *args: Positional arguments for the playbook call context
            **kwargs: Keyword arguments for the playbook call context

        Returns:
            Description with placeholders resolved
        """
        if not description or "{" not in description:
            return description

        try:
            # Create a PlaybookCall for context
            call = PlaybookCall(self.playbook.name, list(args), kwargs)
            context = ExpressionContext(self.agent, self.agent.state, call)
            return await resolve_description_placeholders(description, context)
        except Exception as e:
            logger = logging.getLogger(__name__)
            logger.error(f"Failed to resolve description placeholders: {e}")
            return description

    @abstractmethod
    async def execute(self, *args, **kwargs) -> Any:
        """Execute the playbook with the given strategy.

        Args:
            *args: Positional arguments for the playbook
            **kwargs: Keyword arguments for the playbook

        Returns:
            The execution result
        """
        pass



"""
Meeting system for multi-agent coordination.

This module provides classes and utilities for managing meetings between agents,
including meeting lifecycle management, message routing, and participant coordination.
"""

from .meeting import (
    JoinedMeeting,
    Meeting,
    MeetingInvitation,
    MeetingInvitationStatus,
    Message,
)
from .meeting_manager import MeetingManager

# from .meeting_message_handler import MeetingMessageHandler
from .meeting_registry import MeetingRegistry

__all__ = [
    "Meeting",
    "MeetingInvitation",
    "MeetingInvitationStatus",
    "Message",
    "JoinedMeeting",
    "MeetingManager",
    # "MeetingMessageHandler",
    "MeetingRegistry",
]



"""Traditional playbook execution with defined steps."""

import logging
from typing import TYPE_CHECKING, Any, List

from ..constants import EXECUTION_FINISHED
from ..debug.debug_handler import DebugHandler, NoOpDebugHandler
from ..exceptions import ExecutionFinished
from ..interpreter_prompt import InterpreterPrompt
from ..llm_messages import AssistantResponseLLMMessage, PlaybookImplementationLLMMessage
from ..llm_response import LLMResponse
from ..playbook_call import PlaybookCall
from ..session_log import SessionLogItemLevel, SessionLogItemMessage
from ..utils.expression_engine import (
    ExpressionContext,
    resolve_description_placeholders,
    update_description_in_markdown,
)
from ..utils.llm_config import LLMConfig
from ..utils.llm_helper import get_completion
from ..utils.spec_utils import SpecUtils
from .base import LLMExecution

if TYPE_CHECKING:
    from ..agents.base_agent import Agent
    from ..playbook.llm_playbook import LLMPlaybook

logger = logging.getLogger(__name__)


class PlaybookLLMExecution(LLMExecution):
    """Playbook execution with defined steps.

    This is the core playbook mode - structured natural language functions
    with explicit steps that execute on LLMs. Aligns with ### Steps sections.

    This implements the unified execution strategy that replaced MarkdownPlaybookExecution.
    """

    def __init__(self, agent: "Agent", playbook: "LLMPlaybook"):
        """Initialize playbook execution.

        Args:
            agent: The agent executing the playbook
            playbook: The LLM playbook to execute
        """
        super().__init__(agent, playbook)

        # Initialize debug handler based on whether debug server is available
        # debug(
        #     f"[DEBUG] Initializing debug handler - agent has program: {hasattr(agent, 'program')}, program exists: {agent.program if hasattr(agent, 'program') else 'No'}, debug server exists: {agent.program._debug_server if hasattr(agent, 'program') and agent.program else 'No'}",
        # )
        if hasattr(agent, "program") and agent.program and agent.program._debug_server:
            # Check if debug server already has a debug handler
            if (
                hasattr(agent.program._debug_server, "debug_handler")
                and agent.program._debug_server.debug_handler
            ):
                # Use the existing debug handler from the debug server
                # debug(
                #     "[DEBUG] Using existing DebugHandler from debug server",
                # )
                self.debug_handler = agent.program._debug_server.debug_handler
            else:
                # Create new debug handler and connect it
                # debug(
                #     "[DEBUG] Creating new DebugHandler and connecting to debug server",
                # )
                self.debug_handler = DebugHandler(agent.program._debug_server)
                # Store reference in debug server for bidirectional communication
                agent.program._debug_server.debug_handler = self.debug_handler
                # debug(
                #     "[DEBUG] New debug handler connected to debug server",
                # )
        else:
            self.debug_handler = NoOpDebugHandler()

    async def pre_execute(self, call: PlaybookCall) -> None:
        llm_message = []
        markdown_for_llm = self.playbook.markdown  # Default to original markdown

        # Resolve description placeholders if present
        if self.playbook.description and "{" in self.playbook.description:
            context = ExpressionContext(
                agent=self.agent, state=self.agent.state, call=call
            )
            resolved_description = await resolve_description_placeholders(
                self.playbook.description, context
            )

            markdown_for_llm = update_description_in_markdown(
                self.playbook.markdown, resolved_description
            )

        llm_message.append(
            f"{self.playbook.name} playbook implementation:\n\n```md\n{markdown_for_llm}\n```"
        )

        # Add a cached message whenever we add a stack frame
        llm_message.append("Executing " + str(call))

        # Create a PlaybookImplementationLLMMessage for semantic clarity
        playbook_impl_msg = PlaybookImplementationLLMMessage(
            content="\n\n".join(llm_message), playbook_name=self.playbook.name
        )

        # Add the message object directly to the call stack
        self.agent.state.call_stack.add_llm_message(playbook_impl_msg)

    async def execute(self, *args, **kwargs) -> Any:
        """Execute the playbook with traditional step-by-step logic."""
        done = False
        return_value = None

        call = PlaybookCall(self.playbook.name, args, kwargs)
        await self.pre_execute(call)

        instruction = f"Execute {str(call)} from step 01. Refer to {self.playbook.name} playbook implementation above."
        artifacts_to_load = []

        while not done:
            if self.agent.program.execution_finished:
                break

            description_paragraph = self.agent.description.split("\n\n")[0]
            llm_response = await LLMResponse.create(
                await self.make_llm_call(
                    instruction=instruction,
                    agent_instructions=f"Remember: You are {str(self.agent)}. {description_paragraph}",
                    artifacts_to_load=artifacts_to_load,
                ),
                event_bus=self.agent.state.event_bus,
                agent=self.agent,
            )

            # Create an AssistantResponseLLMMessage for semantic clarity
            llm_response_msg = AssistantResponseLLMMessage(llm_response.response)

            # Add the message object directly to the call stack
            self.agent.state.call_stack.add_llm_message(llm_response_msg)

            artifacts_to_load = []

            all_steps = []
            for line in llm_response.lines:
                for step in line.steps:
                    all_steps.append(step)
            next_steps = {}
            for i in range(len(all_steps)):
                if i == len(all_steps) - 1:
                    next_steps[all_steps[i]] = all_steps[i]
                else:
                    next_steps[all_steps[i]] = all_steps[i + 1]

            for line in llm_response.lines:
                if self.agent.program.execution_finished:
                    break

                if "`SaveArtifact(" not in line.text:
                    for step in line.steps:
                        if step.step:
                            self.agent.state.session_log.append(
                                SessionLogItemMessage(
                                    f"{self.playbook.name}:{step.step.raw_text}"
                                ),
                                level=SessionLogItemLevel.HIGH,
                            )
                    self.agent.state.session_log.append(
                        SessionLogItemMessage(line.text),
                        level=SessionLogItemLevel.LOW,
                    )

                # Process steps but only call handle_execution_start once per line
                for i in range(len(line.steps)):
                    # debug("Execution step", step_index=i, step=str(line.steps[i]))
                    step = line.steps[i]
                    if i == len(line.steps) - 1:
                        next_step = step.copy()
                        # debug("Next step prepared", next_step=str(next_step))
                    else:
                        next_step = step

                    self.agent.state.call_stack.advance_instruction_pointer(next_step)

                # Replace the current call stack frame with the last executed step
                if line.steps:
                    # last_step = line.steps[-1].copy()

                    # debug(
                    #     "Call handle_step",
                    #     last_step=str(last_step),
                    #     next_step=str(next_step),
                    # )

                    for step in line.steps:
                        # debug(f"Agent {self.agent.id} pause if needed on step {step}")
                        await self.debug_handler.pause_if_needed(
                            instruction_pointer=step,
                            agent_id=self.agent.id,
                        )
                        # debug(
                        #     f"Agent {self.agent.id} pause if needed on step {step} done"
                        # )

                # Update variables
                if len(line.vars) > 0:
                    self.agent.state.variables.update(line.vars)

                # Execute playbook calls
                if line.playbook_calls:
                    for playbook_call in line.playbook_calls:
                        if self.agent.program.execution_finished:
                            break

                        # debug("Playbook call", playbook_call=str(playbook_call))
                        if playbook_call.playbook_klass == "Return":
                            if playbook_call.args:
                                return_value = playbook_call.args[0]
                        elif playbook_call.playbook_klass == "LoadArtifact":
                            artifacts_to_load.append(playbook_call.args[0])
                        else:
                            if playbook_call.playbook_klass == "Say":
                                playbook_call.kwargs["already_streamed"] = True
                            await self.agent.execute_playbook(
                                playbook_call.playbook_klass,
                                playbook_call.args,
                                playbook_call.kwargs,
                            )

                # Return value
                if line.return_value:
                    return_value = line.return_value
                    str_return_value = str(return_value)
                    if (
                        str_return_value.startswith("$")
                        and str_return_value in self.agent.state.variables
                    ):
                        return_value = self.agent.state.variables[
                            str_return_value
                        ].value

                # Wait for external event
                if line.wait_for_user_input:
                    await self.agent.WaitForMessage("human")
                elif line.wait_for_agent_input:
                    target_agent_id = self._resolve_yld_target(
                        line.wait_for_agent_target
                    )
                    if target_agent_id:
                        # Check if this is a meeting target
                        if SpecUtils.is_meeting_spec(target_agent_id):
                            meeting_id = SpecUtils.extract_meeting_id(target_agent_id)
                            if meeting_id == "current":
                                meeting_id = (
                                    self.agent.state.call_stack.peek().meeting_id
                                )
                            await self.agent.WaitForMessage(f"meeting {meeting_id}")
                        else:
                            await self.agent.WaitForMessage(target_agent_id)
                elif line.playbook_finished:
                    done = True

                # Raise an exception if line.finished is true
                if line.exit_program:
                    raise ExecutionFinished(EXECUTION_FINISHED)

            # Update instruction
            instruction = []
            for loaded_artifact in artifacts_to_load:
                instruction.append(f"Loaded Artifact[{loaded_artifact}]")
            top_of_stack = self.agent.state.call_stack.peek()
            instruction.append(
                f"{str(top_of_stack)} was executed - "
                f"continue execution. Refer to {top_of_stack.instruction_pointer.playbook} playbook implementation above."
            )

            instruction = "\n".join(instruction)

        if self.agent.program.execution_finished:
            return EXECUTION_FINISHED

        return return_value

    async def make_llm_call(
        self,
        instruction: str,
        agent_instructions: str,
        artifacts_to_load: List[str] = [],
    ):
        """Make an LLM call for playbook execution."""

        prompt = InterpreterPrompt(
            self.agent.state,
            self.agent.playbooks,
            self.playbook,
            instruction=instruction,
            agent_instructions=agent_instructions,
            artifacts_to_load=artifacts_to_load,
            agent_information=self.agent.get_compact_information(),
            other_agent_klasses_information=self.agent.other_agent_klasses_information(),
            trigger_instructions=self.agent.all_trigger_instructions(),
        )

        # Use streaming to handle Say() calls progressively
        return await self._stream_llm_response(prompt)

    async def _stream_llm_response(self, prompt):
        """Stream LLM response and handle Say() calls progressively."""
        buffer = ""
        in_say_call = False
        current_say_content = ""
        say_start_pos = 0
        say_recipient = ""
        processed_up_to = 0  # Track how much of buffer we've already processed

        for chunk in get_completion(
            messages=prompt.messages,
            llm_config=LLMConfig(),
            stream=True,
            json_mode=False,
            langfuse_span=self.agent.state.call_stack.peek().langfuse_span,
        ):
            buffer += chunk

            # Only look for new Say() calls in the unprocessed part of the buffer
            if not in_say_call:
                say_pattern = '`Say("'
                say_match_pos = buffer.find(say_pattern, processed_up_to)
                if say_match_pos != -1:
                    # Found potential Say call - now we need to extract the recipient
                    recipient_start = say_match_pos + len(say_pattern)

                    # Look for the end of the recipient (first argument)
                    recipient_end_pattern = '", "'
                    recipient_end_pos = buffer.find(
                        recipient_end_pattern, recipient_start
                    )

                    if recipient_end_pos != -1:
                        # Extract the recipient
                        say_recipient = buffer[recipient_start:recipient_end_pos]

                        # Only start streaming if recipient is user, human, or Human
                        if say_recipient.lower() in ["user", "human"]:
                            in_say_call = True
                            say_start_pos = recipient_end_pos + len(
                                recipient_end_pattern
                            )  # Position after recipient and ", "
                            current_say_content = ""
                            processed_up_to = say_start_pos
                            await self.agent.start_streaming_say(say_recipient)
                        else:
                            # Not a user/human recipient, skip streaming for this Say call
                            processed_up_to = recipient_end_pos + len(
                                recipient_end_pattern
                            )
                    else:
                        # Haven't found the end of recipient yet, continue processing
                        pass

            # Stream Say content if we're in a call
            if in_say_call:
                # Look for the end of the Say call
                end_pattern = '")'
                end_pos = buffer.find(end_pattern, say_start_pos)
                if end_pos != -1:
                    # Found end - extract final content and complete
                    final_content = buffer[say_start_pos:end_pos]
                    if len(final_content) > len(current_say_content):
                        new_content = final_content[len(current_say_content) :]
                        if new_content:
                            await self.agent.stream_say_update(new_content)

                    await self.agent.complete_streaming_say()
                    in_say_call = False
                    current_say_content = ""
                    say_recipient = ""
                    processed_up_to = end_pos + len(end_pattern)
                else:
                    # Still streaming - extract new content since last update
                    # Only look at content between say_start_pos and end of buffer
                    # but make sure we don't include the closing quote if it's there
                    available_content = buffer[say_start_pos:]

                    # If we see the closing quote, don't include it in streaming
                    if available_content.endswith('")'):
                        available_content = available_content[:-2]  # Remove ")
                    elif available_content.endswith('"'):
                        available_content = available_content[:-1]  # Remove just "

                    # Don't stream if it ends with escape character (incomplete)
                    if not available_content.endswith("\\"):
                        if len(available_content) > len(current_say_content):
                            new_content = available_content[len(current_say_content) :]
                            current_say_content = available_content

                            if new_content:
                                await self.agent.stream_say_update(new_content)

        # If we ended while still in a Say call, complete it
        if in_say_call:
            await self.agent.complete_streaming_say()

        return buffer

    def _resolve_yld_target(self, target: str) -> str:
        """Resolve a YLD target to an agent ID.

        Args:
            target: The YLD target specification

        Returns:
            Resolved agent ID or None if target couldn't be resolved
        """
        if not target:
            return None

        # Use the unified target resolver with no fallback for YLD
        # (YLD should be explicit about what it's waiting for)
        return self.agent.resolve_target(target, allow_fallback=False)



"""Raw LLM call execution without loops or structure."""

from typing import TYPE_CHECKING, Any, List

from ..enums import LLMMessageType
from ..events import PlaybookEndEvent, PlaybookStartEvent
from ..llm_messages import AssistantResponseLLMMessage, UserInputLLMMessage
from ..playbook_call import PlaybookCall
from ..utils.expression_engine import (
    ExpressionContext,
    resolve_description_placeholders,
)
from ..utils.llm_config import LLMConfig
from ..utils.llm_helper import get_completion
from .base import LLMExecution

if TYPE_CHECKING:
    pass


class RawLLMExecution(LLMExecution):
    """Raw LLM call execution without loops or structure.

    This mode:
    - Makes ONE LLM call
    - No loops or iterations
    - No structured steps
    - Direct prompt → response
    """

    async def execute(self, *args, **kwargs) -> Any:
        """Execute with a raw LLM call."""
        # Note: Call stack management is handled by the agent's execute_playbook method
        # No need to push/pop here as it would create double management

        # Publish playbook start event
        self.agent.state.event_bus.publish(
            PlaybookStartEvent(
                agent_id=self.agent.id, session_id="", playbook=self.playbook.name
            )
        )

        # Build the prompt
        messages = await self._build_prompt(*args, **kwargs)

        # Make single LLM call
        response = await self._get_llm_response(messages)

        # Parse and return the response
        result = self._parse_response(response)

        # Publish playbook end event
        call_stack_depth = len(self.agent.state.call_stack.frames)
        self.agent.state.event_bus.publish(
            PlaybookEndEvent(
                agent_id=self.agent.id,
                session_id="",
                playbook=self.playbook.name,
                return_value=result,
                call_stack_depth=call_stack_depth,
            )
        )

        return result

    async def _build_prompt(self, *args, **kwargs) -> str:
        call = PlaybookCall(self.playbook.name, args, kwargs)

        context = ExpressionContext(agent=self.agent, state=self.agent.state, call=call)
        resolved_description = await resolve_description_placeholders(
            self.playbook.description, context
        )

        stack_frame = self.agent.state.call_stack.peek()
        # Get file load messages from the call stack
        messages = [
            msg.to_full_message()
            for msg in stack_frame.llm_messages
            if msg.type == LLMMessageType.FILE_LOAD
        ]
        messages.append(UserInputLLMMessage(resolved_description).to_full_message())
        return messages

    async def _get_llm_response(self, messages: List[dict]) -> str:
        """Get response from LLM."""
        # Get completion
        response_generator = get_completion(
            messages=messages,
            llm_config=LLMConfig(),
            stream=False,
            json_mode=False,
            langfuse_span=self.agent.state.call_stack.peek().langfuse_span,
        )

        response = next(response_generator)

        # Cache the response
        response_msg = AssistantResponseLLMMessage(response)  # cached=True by default
        self.agent.state.call_stack.add_llm_message(response_msg)

        return response

    def _parse_response(self, response: str) -> Any:
        """Parse the LLM response.

        For raw mode, we return the response as-is.
        In the future, this could be enhanced to parse structured outputs.
        """
        return response.strip()



from typing import Any, Callable, List, Optional, Union


def playbook_decorator(
    func_or_triggers: Optional[Union[Callable, List[str]]] = None,
    **kwargs,
) -> Union[Callable, Any]:
    """
    A decorator that marks a function as a playbook. It sets the ``__is_playbook__``
    flag to ``True`` and populates ``__triggers__`` and ``__metadata__`` attributes
    based on the provided arguments. No wrapper function is created; the original
    function is returned unchanged after validation.

    Both synchronous and asynchronous functions are supported. The PythonPlaybook
    execution layer will handle the appropriate calling convention.

    Args:
        func_or_triggers: Either the function to decorate or a list of trigger strings
        triggers: A list of trigger strings when used in the form @playbook(triggers=[...])
        **kwargs: Additional metadata options to store in __metadata__

    Returns:
        The decorated function with __is_playbook__ attribute set to True

    Raises:
        TypeError: If the decorated object is not a callable function
    """
    # Case 1: @playbook used directly (no arguments)
    if callable(func_or_triggers):
        func = func_or_triggers
        # Validate function signature
        if not callable(func):
            raise TypeError("Playbook decorator can only be applied to functions")

        # Store playbook metadata on the function
        func.__is_playbook__ = True
        func.__triggers__ = []
        func.__metadata__ = {}
        return func

    # Case 2: @playbook(triggers=[...]) or @playbook([...]) or @playbook(public=True, etc.)
    else:
        # If triggers is None, assume func_or_triggers is the triggers list
        def decorator(func: Callable) -> Callable:
            # Validate function signature
            if not callable(func):
                raise TypeError("Playbook decorator can only be applied to functions")

            # Store playbook metadata on the function
            func.__is_playbook__ = True
            func.__triggers__ = kwargs.get("triggers", [])

            # Store all kwargs except 'triggers' as metadata
            func.__metadata__ = {k: v for k, v in kwargs.items() if k != "triggers"}

            return func

        return decorator



import uuid
from functools import reduce
from typing import List

import frontmatter

from .compiler import Compiler, FileCompilationResult, FileCompilationSpec
from .event_bus import EventBus
from .loader import Loader
from .logging_setup import configure_logging
from .program import Program
from .utils.llm_config import LLMConfig

# Configure logging at module import
configure_logging()


class Playbooks:
    def __init__(
        self,
        program_paths: List[str],
        llm_config: LLMConfig = None,
        session_id: str = None,
    ):
        self.program_paths = program_paths
        if llm_config is None:
            self.llm_config = LLMConfig()
        else:
            self.llm_config = llm_config.copy()
        self.session_id = session_id or str(uuid.uuid4())

        # Load files
        program_file_tuples = Loader.read_program_files(program_paths)
        self.program_files = [
            FileCompilationSpec(file_path=fp, content=content, is_compiled=is_comp)
            for fp, content, is_comp in program_file_tuples
        ]
        self.program_content = "\n\n".join(
            reduce(
                lambda content, item: content + [item.content], self.program_files, []
            )
        )

        # Check if all files are already compiled (.pbasm files)
        all_files_compiled = all(
            file_spec.is_compiled for file_spec in self.program_files
        )

        if all_files_compiled:
            # Skip compilation - convert FileCompilationSpec to FileCompilationResult directly
            self.compiled_program_files = []
            for file_spec in self.program_files:
                fm_data = frontmatter.loads(file_spec.content)
                self.compiled_program_files.append(
                    FileCompilationResult(
                        file_path=file_spec.file_path,
                        frontmatter_dict=fm_data.metadata,
                        content=fm_data.content,
                        is_compiled=True,
                        compiled_file_path=file_spec.file_path,
                    )
                )
        else:
            # Some files need compilation
            compiler = Compiler(self.llm_config)
            self.compiled_program_files = compiler.process_files(self.program_files)

        # Extract and apply frontmatter from all files (.pb and .pbasm)
        self.program_metadata = {}
        # compiled_content = []
        for i, result in enumerate(self.compiled_program_files):
            if result.frontmatter_dict:
                # Check for duplicate attributes
                for key, value in result.frontmatter_dict.items():
                    if key in self.program_metadata:
                        raise ValueError(
                            f"Duplicate frontmatter attribute '{key}' found in {result.file_path}. "
                            f"Previously defined with value: {self.program_metadata[key]}"
                        )
                    self.program_metadata[key] = value

            # compiled_content.append(file_content)

        # Compiled agents without frontmatter
        # self.compiled_program_content = "\n\n".join(compiled_content)

        # Apply program metadata
        self._apply_program_metadata()

        self.event_bus = EventBus(self.session_id)
        compiled_program_paths = [
            result.compiled_file_path for result in self.compiled_program_files
        ]

        self.program = Program(
            event_bus=self.event_bus,
            program_paths=self.program_paths,
            compiled_program_paths=compiled_program_paths,
            metadata=self.program_metadata,
        )

    async def initialize(self):
        await self.program.initialize()

    async def begin(self):
        await self.program.begin()

    def _apply_program_metadata(self):
        """Apply program-level metadata from frontmatter."""
        for key, value in self.program_metadata.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def get_agent_errors(self):
        """Get a list of all agent errors that have occurred.

        This method provides visibility into agent failures for framework builders
        and playbook authors using the Playbooks class programmatically.

        Returns:
            List of error dictionaries with agent_id, error, and error_type
        """
        if not self.program:
            return []
        return self.program.get_agent_errors()

    def has_agent_errors(self) -> bool:
        """Check if any agents have had errors.

        Returns:
            True if any agent has experienced an error during execution
        """
        if not self.program:
            return False
        return self.program.has_agent_errors()

    def check_execution_health(self) -> dict:
        """Get a comprehensive health check of the playbook execution.

        Returns:
            Dictionary with execution status, error count, and error details
        """
        if not self.program:
            return {
                "status": "not_initialized",
                "has_errors": False,
                "error_count": 0,
                "errors": [],
            }

        errors = self.get_agent_errors()
        return {
            "status": "finished" if self.program.execution_finished else "running",
            "has_errors": len(errors) > 0,
            "error_count": len(errors),
            "errors": errors,
            "execution_finished": self.program.execution_finished,
        }



from typing import List, Optional


class PlaybookTrigger:
    """Represents a trigger that can start a playbook."""

    def __init__(
        self,
        playbook_klass: str,
        playbook_signature: str,
        trigger: str,
        source_line_number: Optional[int] = None,
    ):
        """Initialize a PlaybookTrigger.

        Args:
            playbook_klass: The class name of the playbook.
            playbook_signature: The signature of the playbook function.
            trigger: The trigger string.
            source_line_number: The line number in the source markdown where this
                trigger is defined.
        """
        self.playbook_klass = playbook_klass
        self.playbook_signature = playbook_signature
        self.trigger = trigger
        self.source_line_number = source_line_number
        # Example text: "01:BGN When the agent starts running"
        self.trigger_name = self.trigger.split(" ")[0]
        self.trigger_description = " ".join(self.trigger.split(" ")[1:])
        self.is_begin = "BGN" in self.trigger_name

    def __str__(self) -> str:
        """Return a string representation of the trigger."""
        return self.trigger_instruction()

    def trigger_instruction(self, namespace: Optional[str] = None) -> str:
        if not namespace:
            namespace = ""
        else:
            namespace = namespace + "."

        signature = namespace + self.playbook_signature.split(" ->")[0]
        return f'- {self.trigger_description}, `Trigger["{namespace}{self.playbook_klass}:{self.trigger_name}"]` by enqueuing `{signature}`'


class PlaybookTriggers:
    """Collection of triggers for a playbook."""

    def __init__(
        self,
        playbook_klass: str,
        playbook_signature: str,
        triggers: List[str],
        trigger_line_numbers: Optional[List[Optional[int]]] = None,
        source_line_number: Optional[int] = None,
    ):
        """Initialize a PlaybookTriggers collection.

        Args:
            playbook_klass: The class name of the playbook.
            playbook_signature: The signature of the playbook function.
            triggers: List of trigger strings.
            trigger_line_numbers: List of line numbers for each trigger.
            source_line_number: The line number in the source markdown where this
                triggers section is defined.
        """
        self.playbook_klass = playbook_klass
        self.playbook_signature = playbook_signature
        self.source_line_number = source_line_number

        if trigger_line_numbers is None:
            trigger_line_numbers = [None] * len(triggers)

        self.triggers = [
            PlaybookTrigger(
                playbook_klass=self.playbook_klass,
                playbook_signature=self.playbook_signature,
                trigger=trigger,
                source_line_number=line_num,
            )
            for trigger, line_num in zip(triggers, trigger_line_numbers)
        ]



"""Agent registry for managing agent class registration and discovery."""

from typing import Dict, List, Optional, Type


class AgentClassRegistry:
    """Registry for agent classes that can create agent instances based on configuration."""

    _agents: List[Dict] = []

    @classmethod
    def register(cls, agent_class: Type, priority: int = 0) -> None:
        """Register an agent class with the registry.

        Args:
            agent_class: The agent class to register
            priority: Priority for conflict resolution (higher = more priority)
        """
        # Remove existing registration if present
        cls._agents = [a for a in cls._agents if a["class"] != agent_class]

        # Add new registration
        cls._agents.append({"class": agent_class, "priority": priority})

        # Sort by priority (highest first)
        cls._agents.sort(key=lambda x: x["priority"], reverse=True)

    @classmethod
    def find_agent_class(cls, metadata: Dict) -> Optional[Type]:
        """Find the appropriate agent class for the given metadata.

        Args:
            metadata: Agent metadata from playbook

        Returns:
            Agent class that can handle the metadata, or None if none found
        """
        for agent_info in cls._agents:
            agent_class = agent_info["class"]
            if hasattr(agent_class, "can_handle") and agent_class.can_handle(metadata):
                return agent_class
        return None

    @classmethod
    def get_registered_agents(cls) -> List[Type]:
        """Get list of all registered agent classes."""
        return [a["class"] for a in cls._agents]
